"""Defines types for describing different kinds of mass spectrometry
data files and their contents, and a database of controlled vocabulary
terms for them.
"""

import os
import hashlib
import warnings

from collections import MutableMapping

from six import string_types as basestring

from .cv import Term, TermSet

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = OSError


class IDFormat(Term):
    """Describes a named spectrum identifier format, either
    using a controlled-vocabulary term or user-defined name.

    A :class:`IDFormat` is equal to its name and its controlled
    vocabulary identifier.
    """
    pass


class FileFormat(Term):
    """Describes a named mass spectrometry data file format, either
    using a controlled-vocabulary term or user-defined name.

    A :class:`FileFormat` is equal to its name and its controlled
    vocabulary identifier.
    """
    pass


class FileContent(Term):
    """Describes a named mass spectrometry data file content type,
    either using a controlled-vocabulary term or user-defined name.

    A :class:`FileContent` is equal to its name and its controlled
    vocabulary identifier.
    """
    pass


id_formats = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('native spectrum identifier format',
#             "id_formats", term_cls_name="IDFormat", writer=cog.out)
# ]]]
id_formats = TermSet([
    IDFormat(u'Agilent MassHunter nativeID format', u'MS:1001508',
             (u'Native format defined by scanId=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Shimadzu Biotech QTOF nativeID format', u'MS:1002898',
             (u'Native format defined by scan=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'spectrum from database string nativeID format', u'MS:1001532',
             (u'Native format defined by databasekey=xsd:string.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format', u'spectra data details', u'search input details']),
    IDFormat(u'spectrum from ProteinScape database nativeID format', u'MS:1001531',
             (u'Native format defined by databasekey=xsd:long.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format', u'spectra data details', u'search input details']),
    IDFormat(u'Bruker U2 nativeID format', u'MS:1000823',
             (u'Native format defined by declaration=xsd:nonNegativeInteger'
              u'collection=xsd:nonNegativeInteger'
              u'scan=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'no nativeID format', u'MS:1000824',
             (u'No nativeID format indicates that the file tagged with this'
              u'term does not contain spectra that can have a nativeID'
              u'format.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Mascot query number', u'MS:1001528',
             (u'Native format defined by query=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format', u'spectrum identification result details']),
    IDFormat(u'spectrum from database integer nativeID format', u'MS:1001526',
             (u'Native format defined by databasekey=xsd:long.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Shimadzu Biotech nativeID format', u'MS:1000929',
             (u'Native format defined by source=xsd:string'
              u'start=xsd:nonNegativeInteger end=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'UIMF nativeID format', u'MS:1002532',
             (u'Native format defined by frame=xsd:nonNegativeInteger'
              u'scan=xsd:nonNegativeInteger'
              u'frameType=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Bruker TDF nativeID format', u'MS:1002818',
             (u'Native format defined by frame=xsd:nonNegativeInteger'
              u'scan=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Bruker Container nativeID format', u'MS:1002303',
             (u'Native identifier (UUID).'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'SCIEX TOF/TOF nativeID format', u'MS:1001480',
             (u'Native format defined by jobRun=xsd:nonNegativeInteger'
              u'spotLabel=xsd:string spectrum=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Thermo nativeID format', u'MS:1000768',
             (u'Native format defined by'
              u'controllerType=xsd:nonNegativeInteger'
              u'controllerNumber=xsd:positiveInteger'
              u'scan=xsd:positiveInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Waters nativeID format', u'MS:1000769',
             (u'Native format defined by function=xsd:positiveInteger'
              u'process=xsd:nonNegativeInteger scan=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Scaffold nativeID format', u'MS:1001562',
             (u'Scaffold native ID format.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Bruker FID nativeID format', u'MS:1000773',
             (u'Native format defined by file=xsd:IDREF.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Bruker BAF nativeID format', u'MS:1000772',
             (u'Native format defined by scan=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'Bruker/Agilent YEP nativeID format', u'MS:1000771',
             (u'Native format defined by scan=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'WIFF nativeID format', u'MS:1000770',
             (u'Native format defined by sample=xsd:nonNegativeInteger'
              u'period=xsd:nonNegativeInteger cycle=xsd:nonNegativeInteger'
              u'experiment=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'spectrum identifier nativeID format', u'MS:1000777',
             (u'Native format defined by spectrum=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'scan number only nativeID format', u'MS:1000776',
             (u'Native format defined by scan=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'single peak list nativeID format', u'MS:1000775',
             (u'Native format defined by file=xsd:IDREF.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'multiple peak list nativeID format', u'MS:1000774',
             (u'Native format defined by index=xsd:nonNegativeInteger.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
    IDFormat(u'SCIEX TOF/TOF T2D nativeID format', u'MS:1001559',
             (u'Native format defined by file=xsd:IDREF.'),
             'native spectrum identifier format',
             [u'native spectrum identifier format']),
])
# [[[end]]]


file_formats = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('mass spectrometer file format',
#             "file_formats", term_cls_name="FileFormat", writer=cog.out)
# ]]]
file_formats = TermSet([
    FileFormat(u'Agilent MassHunter format', u'MS:1001509',
               (u'A data file format found in an Agilent MassHunter directory'
                u'which contains raw data acquired by an Agilent mass'
                u'spectrometer.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'msalign format', u'MS:1002899',
               (u'msalign file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'SCiLS Lab format', u'MS:1002385',
               (u'SCiLS Lab file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'chrom format', u'MS:1002966',
               (u'The Lipid Data Analyzer native chrom format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Mascot MGF format', u'MS:1001062',
               (u'Mascot MGF file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Andi-MS format', u'MS:1002441',
               (u'AIA Analytical Data Interchange file format for mass'
                u'spectrometry data.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bruker FID format', u'MS:1000825',
               (u'Bruker FID file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Proteinscape spectra', u'MS:1001527',
               (u'Spectra from Bruker/Protagen Proteinscape database.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'MS2 format', u'MS:1001466',
               (u'MS2 file format for MS2 spectral data." [PMID:15317041,'
                u'DOI:10.1002/rcm.1603'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'PerSeptive PKS format', u'MS:1001245',
               (u'PerSeptive peak list file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'SCIEX API III format', u'MS:1001246',
               (u'PE SCIEX peak list file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bruker XML format', u'MS:1001247',
               (u'Bruker data exchange XML format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Shimadzu Biotech database entity', u'MS:1000930',
               (u'Shimadzu Biotech format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Shimadzu Biotech LCD format', u'MS:1003009',
               (u'Shimadzu Biotech LCD file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'mz5 format', u'MS:1001881',
               (u'mz5 file format, modelled after mzML.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'mzMLb format', u'MS:1002838',
               (u'mzMLb file format, mzML encapsulated within HDF5." [PSI:PI'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'UIMF format', u'MS:1002531',
               (u'SQLite-based file format created at Pacific Northwest'
                u'National Lab. It stores an intermediate analysis of ion-'
                u'mobility mass spectrometry data.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bruker TDF format', u'MS:1002817',
               (u'Bruker TDF raw file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Waters raw format', u'MS:1000526',
               (u'Waters data file format found in a Waters RAW directory,'
                u'generated from an MS acquisition.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'feature format', u'MS:1002900',
               (u'TopFD feature file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bruker Container format', u'MS:1002302',
               (u'Bruker Container raw file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Phenyx XML format', u'MS:1001463',
               (u'Phenyx open XML file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'intermediate analysis format', u'file format']),
    FileFormat(u'mzML format', u'MS:1000584',
               (u'Proteomics Standards Inititative mzML file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bioworks SRF format', u'MS:1000742',
               (u'Thermo Finnigan SRF file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'intermediate analysis format', u'file format']),
    FileFormat(u'parameter file', u'MS:1000740',
               (u'Parameter file used to configure the acquisition of raw data'
                u'on the instrument.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Andromeda:apl file format', u'MS:1002996',
               (u'Peak list file format of the Andromeda search engine.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'ISB mzXML format', u'MS:1000566',
               (u'Institute of Systems Biology mzXML file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bruker/Agilent YEP format', u'MS:1000567',
               (u'Bruker/Agilent YEP file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'PSI mzData format', u'MS:1000564',
               (u'Proteomics Standards Inititative mzData file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Micromass PKL format', u'MS:1000565',
               (u'Micromass PKL file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'ABI WIFF format', u'MS:1000562',
               (u'Applied Biosystems WIFF file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Thermo RAW format', u'MS:1000563',
               (u'Thermo Scientific RAW file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'SCIEX TOF/TOF database', u'MS:1001481',
               (u'Applied Biosystems/MDS Analytical Technologies TOF/TOF'
                u'instrument database.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'DTA format', u'MS:1000613',
               (u'SEQUEST DTA file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'ProteinLynx Global Server mass spectrum XML format', u'MS:1000614',
               (u'Peak list file format used by ProteinLynx Global Server.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'text format', u'MS:1001369',
               (u'Simple text file format of \\"m/z [intensity]\\" values for a'
                u'PMF (or single MS2) search.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'SCIEX TOF/TOF T2D format', u'MS:1001560',
               (u'Applied Biosystems/MDS Analytical Technologies TOF/TOF'
                u'instrument export format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'MS1 format', u'MS:1002597',
               (u'MS1 file format for MS1 spectral data." [PMID:15317041'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bruker U2 format', u'MS:1000816',
               (u'Bruker HyStar U2 file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
    FileFormat(u'Bruker BAF format', u'MS:1000815',
               (u'Bruker BAF raw file format.'),
               'mass spectrometer file format',
               [u'mass spectrometer file format', u'file format']),
])
# [[[end]]]


content_keys = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('data file content',
#             "content_keys", term_cls_name="FileContent", writer=cog.out)
# ]]]
content_keys = TermSet([
    FileContent(u'electromagnetic radiation spectrum', u'MS:1000804',
                (u'A plot of the relative intensity of electromagnetic'
                 u'radiation as a function of the wavelength.'),
                'data file content',
                [u'data file content', u'spectrum type']),
    FileContent(u'absorption spectrum', u'MS:1000806',
                (u'A plot of the relative intensity of electromagnetic'
                 u'radiation absorbed by atoms or molecules when excited.'),
                'data file content',
                [u'data file content', u'spectrum type']),
    FileContent(u'emission spectrum', u'MS:1000805',
                (u'A plot of the relative intensity of electromagnetic'
                 u'radiation emitted by atoms or molecules when excited.'),
                'data file content',
                [u'data file content', u'spectrum type']),
    FileContent(u'mass spectrum', u'MS:1000294',
                (u'A plot of the relative abundance of a beam or other'
                 u'collection of ions as a function of the mass-to-charge ratio'
                 u'(m/z).'),
                'data file content',
                [u'data file content', u'spectrum type']),
    FileContent(u'PDA spectrum', u'MS:1000620',
                (u'OBSOLETE Spectrum generated from a photodiode array detector'
                 u'(ultraviolet/visible spectrum).'),
                'data file content',
                [u'data file content', u'spectrum type']),
    FileContent(u'mass chromatogram', u'MS:1000810',
                (u'A plot of the relative abundance of a beam or other'
                 u'collection of ions as a function of the retention time.'),
                'data file content',
                [u'data file content', u'chromatogram type']),
    FileContent(u'electromagnetic radiation chromatogram', u'MS:1000811',
                (u'The measurement of electromagnetic properties as a function'
                 u'of the retention time.'),
                'data file content',
                [u'data file content', u'chromatogram type']),
    FileContent(u'MSn spectrum', u'MS:1000580',
                (u'MSn refers to multi-stage MS2 experiments designed to record'
                 u'product ion spectra where n is the number of product ion'
                 u'stages (progeny ions). For ion traps, sequential MS/MS'
                 u'experiments can be undertaken where n > 2 whereas for a'
                 u'simple triple quadrupole system n=2. Use the term ms level'
                 u'(MS:1000511) for specifying n.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'CRM spectrum', u'MS:1000581',
                (u'Spectrum generated from MSn experiment with three or more'
                 u'stages of m/z separation and in which a particular multi-'
                 u'step reaction path is monitored.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'SIM spectrum', u'MS:1000582',
                (u'Spectrum obtained with the operation of a mass spectrometer'
                 u'in which the abundances of one ion or several ions of'
                 u'specific m/z values are recorded rather than the entire mass'
                 u'spectrum (Selected Ion Monitoring).'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'SRM spectrum', u'MS:1000583',
                (u'Spectrum obtained when data are acquired from specific'
                 u'product ions corresponding to m/z values of selected'
                 u'precursor ions a recorded via two or more stages of mass'
                 u'spectrometry. The precursor/product ion pair is called a'
                 u'transition pair. Data can be obtained for a single'
                 u'transition pair or multiple transition pairs. Multiple time'
                 u'segments of different transition pairs can exist in a single'
                 u'file. Single precursor ions can have multiple product ions'
                 u'consitituting multiple transition pairs. Selected reaction'
                 u'monitoring can be performed as tandem mass spectrometry in'
                 u'time or tandem mass spectrometry in space.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'e/2 mass spectrum', u'MS:1000328',
                (u'A mass spectrum obtained using a sector mass spectrometer in'
                 u'which the electric sector field E is set to half the value'
                 u'required to transmit the main ion-beam. This spectrum'
                 u'records the signal from doubly charged product ions of'
                 u'charge-stripping reactions.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'constant neutral gain spectrum', u'MS:1000325',
                (u'A spectrum formed of all product ions that have been'
                 u'produced by gain of a pre-selected neutral mass following'
                 u'the reaction with and addition of the gas in a collision'
                 u'cell.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'constant neutral loss spectrum', u'MS:1000326',
                (u'A spectrum formed of all product ions that have been'
                 u'produced with a selected m/z decrement from any precursor'
                 u'ions. The spectrum shown correlates to the precursor ion'
                 u'spectrum. See also neutral loss spectrum.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'charge inversion mass spectrum', u'MS:1000322',
                (u'The measurement of the relative abundance of ions that'
                 u'result from a charge inversion reaction as a function of'
                 u'm/z.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'product ion spectrum', u'MS:1000343',
                (u'OBSOLETE A mass spectrum recorded from any spectrometer in'
                 u'which the appropriate m/z separation scan function is set to'
                 u'record the product ion or ions of selected precursor ions.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'precursor ion spectrum', u'MS:1000341',
                (u'Spectrum generated by scanning precursor m/z while'
                 u'monitoring a fixed product m/z.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'MS1 spectrum', u'MS:1000579',
                (u'Mass spectrum created by a single-stage MS experiment or the'
                 u'first stage of a multi-stage experiment.'),
                'data file content',
                [u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'total ion current chromatogram', u'MS:1000235',
                (u'Chromatogram obtained by plotting the total ion current'
                 u'detected in each of a series of mass spectra recorded as a'
                 u'function of retention time.'),
                'data file content',
                [u'mass chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'selected reaction monitoring chromatogram', u'MS:1001473',
                (u'Chromatogram created by creating an array of the'
                 u'measurements of a selectively monitored reaction at each'
                 u'time point.'),
                'data file content',
                [u'mass chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'selected ion monitoring chromatogram', u'MS:1001472',
                (u'Chromatogram created by creating an array of the'
                 u'measurements of a selectively monitored ion at each time'
                 u'point.'),
                'data file content',
                [u'mass chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'consecutive reaction monitoring chromatogram', u'MS:1001474',
                (u'OBSOLETE Chromatogram created by creating an array of the'
                 u'measurements of a series of monitored reactions at each time'
                 u'point.'),
                'data file content',
                [u'mass chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'basepeak chromatogram', u'MS:1000628',
                (u'Chromatogram created by creating an array of the most'
                 u'intense peaks at each time point.'),
                'data file content',
                [u'mass chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'selected ion current chromatogram', u'MS:1000627',
                (u'Chromatogram created by creating an array of the'
                 u'measurements of a specific single ion current at each time'
                 u'point.'),
                'data file content',
                [u'mass chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'absorption chromatogram', u'MS:1000812',
                (u'The measurement of light absorbed by the sample as a'
                 u'function of the retention time.'),
                'data file content',
                [u'electromagnetic radiation chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'emission chromatogram', u'MS:1000813',
                (u'The measurement of light emitted by the sample as a function'
                 u'of the retention time.'),
                'data file content',
                [u'electromagnetic radiation chromatogram', u'data file content', u'chromatogram type']),
    FileContent(u'time-delayed fragmentation spectrum', u'MS:1000790',
                (u'MSn spectrum in which the product ions are collected after a'
                 u'time delay, which allows the observation of lower energy'
                 u'fragmentation processes after precursor ion activation.'),
                'data file content',
                [u'MSn spectrum', u'mass spectrum', u'data file content', u'spectrum type']),
    FileContent(u'enhanced multiply charged spectrum', u'MS:1000789',
                (u'MS1 spectrum that is enriched in multiply-charged ions'
                 u'compared to singly-charged ions.'),
                'data file content',
                [u'MS1 spectrum', u'mass spectrum', u'data file content', u'spectrum type']),
])
# [[[end]]]


spectrum_representation = []
# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('spectrum representation',
#             "spectrum_representation", term_cls_name="FileContent", writer=cog.out)
# ]]]
spectrum_representation = TermSet([
    FileContent(u'centroid spectrum', u'MS:1000127',
                (u'Processing of profile data to produce spectra that contains'
                 u'discrete peaks of zero width. Often used to reduce the size'
                 u'of dataset.'),
                'spectrum representation',
                [u'spectrum representation']),
    FileContent(u'profile spectrum', u'MS:1000128',
                (u'A profile mass spectrum is created when data is recorded'
                 u'with ion current (counts per second) on one axis and'
                 u'mass/charge ratio on another axis.'),
                'spectrum representation',
                [u'spectrum representation']),
])
# [[[end]]]


content_keys = content_keys + spectrum_representation

id_formats_by_name = {k.name: k for k in id_formats}
file_formats_by_name = {k.name: k for k in file_formats}
content_keys_by_name = {k.name: k for k in content_keys}


MS_MS1_Spectrum = content_keys.get('MS1 spectrum')
MS_MSn_Spectrum = content_keys.get('MSn spectrum')


def id_format(name):
    '''Translate a given name or identifier into a :class:`IDFormat`
    instance.

    If no match is found in the database of known :class:`IDFormat`
    types, a new dummy :class:`IDFormat` is returned with all fields
    set to the value of ``name``

    Returns
    -------
    IDFormat
    '''
    try:
        return id_formats[name]
    except KeyError:
        if name is None:
            return None
        return IDFormat(name, name, name, name, [name])


def file_format(name):
    '''Translate a given name or identifier into a :class:`FileFormat`
    instance.

    If no match is found in the database of known :class:`FileFormat`
    types, a new dummy :class:`FileFormat` is returned with all fields
    set to the value of ``name``

    Returns
    -------
    FileFormat
    '''
    try:
        return file_formats[name]
    except KeyError:
        if name is None:
            return None
        return FileFormat(name, name, name, name, [name])


def content_key(name):
    '''Translate a given name or identifier into a :class:`FileContent`
    instance.

    If no match is found in the database of known :class:`FileContent`
    types, a new dummy :class:`FileContent` is returned with all fields
    set to the value of ``name``

    Returns
    -------
    FileContent
    '''
    try:
        return content_keys[name]
    except KeyError:
        if name is None:
            return None
        return FileContent(name, name, name, name, [name])


class FileInformation(MutableMapping):
    """Describes the type of data found in this file and the
    source files that contributed to it.

    Implements the :class:`MutableMapping` Interface

    Attributes
    ----------
    contents : dict
        A mapping between controlled vocabullary names or user-defined
        names and an optional value. For standard controlled names see
        :data:`content_keys`
    source_files : list of :class:`.SourceFile` objects
        The set of files which either define the current file, or were used
        to create the current file if recorded.
    """

    def __init__(self, contents=None, source_files=None):
        if contents is None:
            contents = {}
        if source_files is None:
            source_files = []
        self.contents = dict(contents)
        self.source_files = list(source_files)

    def add_file(self, source, check=True):
        """Add a new file to :attr:`source_files`

        If ``source`` is a string, it will be interpreted as a path
        and an instance of :class:`SourceFile` will be created using
        :meth:`SourceFile.from_path`. Otherwise, it is assumed to be
        an instance of :class:`SourceFile`.

        Parameters
        ----------
        source : str or :class:`SourceFile`
            Either the path to a file to be added to the source file
            collection, or an instance of :class:`SourceFile`
        check : bool, optional
            Whether or not to check and validate that a path points to
            a real file

        Raises
        ------
        ValueError
            If a path fails to validate as real
        """
        if isinstance(source, basestring):
            source = os.path.realpath(source)
            if check:
                if not os.path.exists(source):
                    raise ValueError(
                        "Source File %r does not exist" % (source,))
            source = SourceFile.from_path(source)
        elif not isinstance(source, SourceFile):
            raise TypeError("Must pass an object of type %r, could not coerce %r" % (SourceFile, type(source)))
        self.source_files.append(source)

    def add_content(self, key, value=None):
        """Adds a new key-value pair to :attr:`contents` with
        an optional value

        Parameters
        ----------
        key : str or :attr:`content`
            The content name, either a CV-term or a user-defined name
        value : object, optional
            The optional value, which should be any type of object whose
            meaning makes sense given the definition of ``key``
        """
        self.contents[key] = value

    def remove_content(self, key):
        """Remove a key from :attr:`content`

        Parameters
        ----------
        key : str or :class:`FileContent`
            The content key to remove
        """
        self.contents.pop(key, None)

    def get_content(self, key):
        '''Retrieve the value of ``key`` from :attr:`contents`.

        This method is aliased to :meth:`__getitem__`

        Parameters
        ----------
        key : str or :class:`FileContent`

        Returns
        -------
        object
        '''
        return self.contents[key]

    def has_content(self, key):
        '''Check if ``key`` is found in :attr:`content`

        Parameters
        ----------
        key : str or :class:`FileContent`

        Returns
        -------
        bool
        '''
        return key in self.contents

    def __getitem__(self, key):
        return self.get_content(key)

    def __contains__(self, key):
        return self.has_content(key)

    def __setitem__(self, key, value):
        self.add_content(key, value)

    def __delitem__(self, key):
        self.remove_content(key)

    def __repr__(self):
        template = "FileInformation(%s, %s)"
        return template % (self.contents, self.source_files)

    def __len__(self):
        return len(self.contents)

    def __iter__(self):
        return iter(self.contents)

    def copy(self):
        '''Create a deep copy of this object

        Returns
        -------
        FileInformation
        '''
        return self.__class__(
            self.contents.copy(), [f.copy() for f in self.source_filesr])


format_parameter_map = {
    "thermo raw": (id_formats_by_name.get("Thermo nativeID format"),
                   file_formats_by_name.get("Thermo RAW format")),
    "agilent d": (id_formats_by_name.get("Agilent MassHunter nativeID format"),
                  file_formats_by_name.get("Agilent MassHunter format")),
    'mgf': (id_formats_by_name.get("no nativeID format"),
            # id_formats_by_name.get("multiple peak list nativeID format"),
            file_formats_by_name.get('Mascot MGF format')),
}


class SourceFile(object):
    """Represents a single raw data file which either defines or contributed data to another
    data file, the "reference file"

    Attributes
    ----------
    file_format : :class:`~.FileFormat`
        The name of a data file format. See :data:`file_formats`
    id : str
        The unique identifier for this file, among files which contributed to
        the reference file
    id_format : :class:`~.IDFormat`
        The name of a formal identifier schema. See :data:`~.id_formats`
    location : str
        The directory path to this file on the machine it was last read on to contribute to
        or define the reference file
    name : str
        The base name of this file
    parameters : dict
        A set of key-value pairs associated with this file, either encoding extra metadata
        annotations, or precomputed hash checksums
    """

    _checksum_translation_map = {
        'sha1': 'sha1',
        'SHA-1': 'sha1',
        'md5': 'md5',
        'MD5': 'md5'
    }

    @classmethod
    def _resolve_checksum_hash_type(cls, hash_type):
        try:
            return cls._checksum_translation_map[hash_type]
        except KeyError:
            try:
                return cls._checksum_translation_map[hash_type.lower().replace("-", '')]
            except KeyError:
                raise KeyError(hash_type)

    @classmethod
    def from_path(cls, path):
        '''Construct a new :class:`SourceFile` from a path to a real file on the local
        file system.

        Parameters
        ----------
        path: str
            The path to the file to describe

        Returns
        -------
        SourceFile
        '''
        path = os.path.realpath(path)
        name = os.path.basename(path)
        location = os.path.dirname(path)
        idfmt, file_fmt = cls.guess_format(path)
        source = cls(name, location, name, idfmt, file_fmt)
        return source

    def __init__(self, name, location, id=None, id_format=None, file_format=None, parameters=None):
        if id is None:
            id = name
        if parameters is None:
            parameters = {}
        self.name = name
        self.location = location.replace("file:///", '')
        self.id = id
        self.parameters = dict(parameters)
        self._clean_parameters()
        self.id_format = id_format or self.infer_id_format_from_paramters()
        self.file_format = file_format or self.infer_file_format_from_parameters()

    @property
    def id_format(self):
        return self._id_format

    @id_format.setter
    def id_format(self, value):
        if value is None:
            self._id_format = None
        else:
            self._id_format = id_format(str(value))

    @property
    def file_format(self):
        return self._file_format

    @file_format.setter
    def file_format(self, value):
        if value is None:
            self._file_format = None
        else:
            self._file_format = file_format(str(value))

    def _clean_parameters(self):
        self.parameters.pop("location", None)
        self.parameters.pop("id", None)
        self.parameters.pop("name", None)

    def infer_id_format_from_paramters(self):
        try:
            fmt = list(set(self.parameters) & set(id_formats))[0]
            self.parameters.pop(fmt)
            return fmt
        except IndexError:
            return None

    def infer_file_format_from_parameters(self):
        try:
            fmt = list(set(self.parameters) & set(file_formats))[0]
            self.parameters.pop(fmt)
            return fmt
        except IndexError:
            return None

    @property
    def path(self):
        return os.path.join(self.location, self.name)

    def is_resolvable(self):
        return os.path.exists(self.path)

    @staticmethod
    def guess_format(path):
        if not os.path.exists(path):
            return None, None
        if os.path.isdir(path):
            if os.path.exists(os.path.join(path, 'AcqData')):
                return format_parameter_map['agilent d']

        parts = os.path.splitext(path)
        if len(parts) > 1:
            ext = parts[1]
            if ext.lower() == '.mzml':
                fmt = "mzML format"
                id_fmt = "no nativeID format"
                hit = False
                with open(path, 'rb') as fh:
                    from ..xml_reader import iterparse_until
                    for sf_tag in iterparse_until(fh, 'sourceFile', 'run'):
                        for param in sf_tag.getchildren():
                            if "nativeID" in param.attrib['name']:
                                id_fmt = param.attrib['name']
                                hit = True
                                break
                        if hit:
                            break
                return id_fmt, fmt
            elif ext.lower() == '.mzxml':
                fmt = "ISB mzXML format"
                id_fmt = "no nativeID format"
                return fmt, id_fmt
            elif ext.lower() == '.mgf':
                fmt = "Mascot MGF format"
                id_fmt = "no nativeID format"
                return fmt, id_fmt
        with open(path, 'rb') as fh:
            lead_bytes = fh.read(30)
            # looking for pattern matching b'\x01\xa1F\x00i\x00n\x00n\x00i\x00g\x00a\x00n\x00'
            decoded = lead_bytes.decode("utf-16")[1:9]
            if decoded == "Finnigan":
                return format_parameter_map['thermo raw']
        return None, None

    def __repr__(self):
        template = "SourceFile(%r, %r, %r, %s, %s%s)"
        if self.parameters:
            tail = ", %r" % self.parameters
        else:
            tail = ''
        return template % (
            self.name, self.location,
            self.id, self.id_format, self.file_format,
            tail
        )

    def _compute_checksum(self, hash_type='sha1', buffer_size=2**16):
        hasher = hashlib.new(hash_type)
        buffer_size = int(buffer_size)
        with open(self.path, 'rb') as fh:
            content_buffer = fh.read(buffer_size)
            while content_buffer:
                hasher.update(content_buffer)
                content_buffer = fh.read(buffer_size)
        return hasher.hexdigest()

    def checksum(self, hash_type='sha1'):
        hash_type = self._resolve_checksum_hash_type(hash_type)
        return self._compute_checksum(hash_type)

    def add_checksum(self, hash_type='sha1'):
        hash_type = self._resolve_checksum_hash_type(hash_type)
        checksum = self.checksum(hash_type)
        if hash_type == 'sha1':
            self.parameters['SHA-1'] = checksum
        elif hash_type == "md5":
            self.parameters['MD5'] = checksum

    def validate_checksum(self):
        if not os.path.exists(self.path):
            FileNotFoundError("%s not found" % (self.path,))
        if 'SHA-1' in self.parameters:
            checksum = self.checksum('sha1')
            return self.parameters['SHA-1'] == checksum
        elif 'MD5' in self.parameters:
            checksum = self.checksum("md5")
            return self.parameters['MD5'] == checksum
        else:
            warnings.warn(
                "%r did not have a reference checksum. Could not validate" % (self,))
            return True

    def has_checksum(self, hash_type=None):
        if hash_type is None:
            return ("SHA-1" in self.parameters) or ("MD5" in self.parameters)
        elif self._resolve_checksum_hash_type(hash_type) == 'sha1':
            return ("SHA-1" in self.parameters)
        elif self._resolve_checksum_hash_type(hash_type) == 'md5':
            return "MD5" in self.parameters

    def __eq__(self, other):
        if other is None:
            return False
        if self.path == other.path:
            if self.is_resolvable() and other.is_resolvable():
                return self.checksum() == other.checksum()
            else:
                if self.is_resolvable():
                    for hash_type in ['SHA-1', 'MD5']:
                        if other.has_checksum(hash_type):
                            return self.checksum(hash_type) == other.parameters[hash_type]
                elif other.is_resolvable():
                    for hash_type in ['SHA-1', 'MD5']:
                        if self.has_checksum(hash_type):
                            return other.checksum(hash_type) == self.parameters[hash_type]
                else:
                    for hash_type in ['SHA-1', 'MD5']:
                        if self.has_checksum(hash_type) and other.has_checksum(hash_type):
                            return other.parameters[hash_type] == self.parameters[hash_type]
        return False

    def __ne__(self, other):
        return not self == other

    def copy(self):
        return self.__class__(self.name, self.location, self.id,
                              self.id_format, self.file_format,
                              parameters=self.parameters.copy())


__all__ = [
    "IDFormat", "FileFormat", "FileContent",
    "id_formats", "file_formats", "content_keys",
    "id_format", "file_format", "content_key",
    "FileInformation", "SourceFile"
]



'''**mzXML** is a standard XML-format for raw mass spectrometry data storage created
by the Institute for Systems Biology, intended to be replaced with **mzML**.
This module provides :class:`MzXMLLoader`, a :class:`~.RandomAccessScanSource`
implementation.

The parser is based on :mod:`pyteomics.mzxml`.
'''
from six import string_types as basestring

import numpy as np
from pyteomics import mzxml
from .common import (
    PrecursorInformation, ScanDataSource, ChargeNotProvided,
    ActivationInformation, IsolationWindow, ScanAcquisitionInformation,
    ScanEventInformation, ScanWindow,
    ComponentGroup, component, InstrumentInformation,
    FileInformation, ScanFileMetadataBase)
from .metadata import data_transformation
from .xml_reader import (
    XMLReaderBase, iterparse_until)


class _MzXMLParser(mzxml.MzXML):
    pass


class _MzXMLMetadataLoader(ScanFileMetadataBase):
    def file_description(self):
        """Read the file provenance from the ``<parentFile>`` tags
        if any are present.

        This returns no information about the file's contents as this
        was not part of the mzXML schema

        Returns
        -------
        FileInformation
            The description of the file's  sources
        """
        file_info = map(self.source._get_info_smart, iterparse_until(self.source, "parentFile", "scan"))
        self.source.reset()
        file_info = list(file_info)
        fi = FileInformation({}, [])
        for parent in file_info:
            path = parent.get("fileName")
            if path is None:
                continue
            path = path.replace("file:///", '')
            fi.add_file(path, check=False)
        return fi

    def instrument_configuration(self):
        """Read the instrument configurations settings from the
        ``<msInstrument>`` elements.

        Returns
        -------
        list of InstrumentConfiguration
            A list of different instrument states that scans may be acquired under
        """
        instrument_configuration = map(self.source._get_info_smart, iterparse_until(
            self.source, "msInstrument", "scan"))
        self.source.reset()
        instrument_configuration = [
            self._convert_instrument(ic) for ic in instrument_configuration
        ]
        return instrument_configuration

    def _convert_instrument(self, configuration):
        try:
            detector = configuration.get('msDetector', {}).get('value')
        except AttributeError:
            detector = None
        if detector is not None:
            detector = component(detector)
        try:
            ionisation = configuration.get('msIonisation', {}).get('value')
        except AttributeError:
            ionisation = None
        if ionisation is not None:
            ionisation = component(ionisation)
        try:
            analyzer = configuration.get('msMassAnalyzer', {}).get('value')
        except AttributeError:
            analyzer = None
        if analyzer is not None:
            analyzer = component(analyzer)
        parts = [
            ComponentGroup("source", [ionisation], 1),
            ComponentGroup("analyzer", [analyzer], 2),
            ComponentGroup("detector", [detector], 3)
        ]
        return InstrumentInformation(configuration.get('msInstrumentID', 1), parts)

    def data_processing(self):
        data_processing = map(self.source._get_info_smart, iterparse_until(
            self.source, "dataProcessing", "scan"))
        self.source.reset()
        operation_groups = []
        for i, group in enumerate(data_processing):
            software_id = group.get("software", {}).get("name", "")
            method_group = data_transformation.ProcessingMethod(order=1, software_id=software_id)
            operations = group.get("processingOperation", [])
            if isinstance(operations, list):
                for op in operations:
                    method_group.add(op)
            else:
                method_group.add(operations)
            operation_groups.append(
                data_transformation.DataProcessingInformation([method_group], id=i))
        return operation_groups


class MzXMLDataInterface(ScanDataSource):
    """Provides implementations of all of the methods needed to implement the
    :class:`ScanDataSource` for mzXML files. Not intended for direct instantiation.
    """
    def _scan_arrays(self, scan):
        """Returns raw data arrays for m/z and intensity

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        mz: np.array
            An array of m/z values for this scan
        intensity: np.array
            An array of intensity values for this scan
        """
        try:
            return (scan['m/z array'], scan["intensity array"])
        except KeyError:
            return np.array([]), np.array([])

    def _precursor_information(self, scan):
        """Returns information about the precursor ion,
        if any, that this scan was derived form.

        Returns `None` if this scan has no precursor ion

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        PrecursorInformation
        """
        try:
            pinfo_dict = scan['precursorMz'][0]
            precursor_scan_id = pinfo_dict.get('precursorScanNum')
            pinfo = PrecursorInformation(
                mz=float(pinfo_dict['precursorMz']),
                intensity=float(pinfo_dict.get('precursorIntensity', 0.0)),
                charge=int(pinfo_dict.get('precursorCharge')) if pinfo_dict.get(
                    'precursorCharge') else ChargeNotProvided,
                precursor_scan_id=precursor_scan_id,
                source=self,
                product_scan_id=self._scan_id(scan))
            return pinfo
        except KeyError:
            return None

    def _scan_title(self, scan):
        """Returns a verbose name for this scan, if one
        was stored in the file. Usually includes both the
        scan's id string, as well as information about the
        original file and format.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        return self._scan_id(scan)

    def _scan_id(self, scan):
        """Returns the scan's id string, a unique
        identifier for this scan in the context of
        the data file it is recordered in

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        return scan["num"]

    def _scan_index(self, scan):
        """Returns the base 0 offset from the start
        of the data file in number of scans to reach
        this scan.

        If the original format does not natively include
        an index value, this value may be computed from
        the byte offset index.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        try:
            if self._scan_index_lookup is None:
                raise ValueError("Index Not Built")
            scan_index = self._scan_index_lookup[self._scan_id(scan)]
            return scan_index
        except KeyError:
            return -1
        except ValueError:
            return -2

    def _ms_level(self, scan):
        """Returns the degree of exponential fragmentation
        used to produce this scan.

        1 refers to a survey scan of unfragmented ions, 2
        refers to a tandem scan derived from an ms level 1
        ion, and so on.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        return int(scan['msLevel'])

    def _scan_time(self, scan):
        """Returns the time in minutes from the start of data
        acquisition to when this scan was acquired.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        float
        """
        try:
            return scan['retentionTime']
        except KeyError:
            return None

    def _is_profile(self, scan):
        """Returns whether the scan contains profile data (`True`)
        or centroided data (`False`).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        bool
        """
        try:
            return not bool(int(scan['centroided']))
        except KeyError:
            return True

    def _polarity(self, scan):
        """Returns whether this scan was acquired in positive mode (+1)
        or negative mode (-1).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        try:
            if scan['polarity'] == '+':
                return 1
            else:
                return -1
        except KeyError:
            return None

    def _activation(self, scan):
        """Returns information about the activation method used to
        produce this scan, if any.

        Returns `None` for MS1 scans

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        ActivationInformation
        """
        try:
            return ActivationInformation(
                scan['precursorMz'][0]['activationMethod'], scan['collisionEnergy'])
        except KeyError:
            return None

    def _isolation_window(self, scan):
        try:
            pinfo_dict = scan['precursorMz'][0]
            target = float(pinfo_dict['precursorMz'])
            width = float(pinfo_dict['windowWideness'])
            lower = width / 2
            upper = width / 2
            return IsolationWindow(lower, target, upper)
        except KeyError:
            try:
                pinfo_dict = scan['precursorMz'][0]
                target = float(pinfo_dict['precursorMz'])
                return IsolationWindow.make_empty(target)
            except KeyError:
                return None

    def _instrument_configuration(self, scan):
        try:
            return self._instrument_config[scan['msInstrumentID']]
        except KeyError:
            return None

    def _acquisition_information(self, scan):
        scan_event = ScanEventInformation(
            scan['retentionTime'],
            window_list=[
                ScanWindow(scan.get("lowMz"), scan.get("highMz"))
            ])
        return ScanAcquisitionInformation("no combination", [scan_event])


class MzXMLLoader(MzXMLDataInterface, XMLReaderBase, _MzXMLMetadataLoader):
    """Reads scans from mzXML files. Provides both iterative and
    random access.

    Attributes
    ----------
    source_file: str
        Path to file to read from.
    source: pyteomics.mzxml.MzXML
        Underlying scan data source
    """

    _parser_cls = _MzXMLParser


    def __init__(self, source_file, use_index=True, **kwargs):
        self.source_file = source_file
        self._source = _MzXMLParser(source_file, read_schema=True, iterative=True,
                                    huge_tree=True, use_index=use_index)
        self.initialize_scan_cache()
        self._use_index = use_index
        self._scan_index_lookup = None
        if self._use_index:
            self._build_scan_index_lookup()
        self._instrument_config = {
            k.id: k for k in self.instrument_configuration()
        }
        self.reset()
        self.make_iterator()

    @property
    def index(self):
        return self._source.index['scan']

    def _get_scan_by_id_raw(self, scan_id):
        return self._source.get_by_id(scan_id, "num")

    def _build_scan_index_lookup(self):
        if not self._use_index:
            raise ValueError("Must index the entire file before sequential indices may computed.")
        index = dict()
        i = 0
        for scan, _offset in self.index.items():
            index[scan] = i
            i += 1
        self._scan_index_lookup = index

    def _validate(self, scan):
        return "m/z array" in scan._data

    def _yield_from_index(self, scan_source, start=None):
        offset_provider = scan_source._offset_index['scan']
        keys = list(offset_provider.keys())
        if start is not None:
            if isinstance(start, basestring):
                start = keys.index(start)
            elif isinstance(start, int):
                start = start
            else:
                raise TypeError("Cannot start from object %r" % start)
        else:
            start = 0
        for key in keys[start:]:
            scan = scan_source.get_by_id(key, "num")
            yield scan



from collections import namedtuple

from ms_deisotope.averagine import mass_charge_ratio, peptide as peptide_averagine
from ms_deisotope.peak_set import Envelope, EnvelopePair


_CoIsolation = namedtuple(
    "CoIsolation", ("neutral_mass", "intensity", "charge"))


class CoIsolation(_CoIsolation):
    '''Records the properties of a co-isolating ion associated with a primary ion.

    Attributes
    ----------
    neutral_mass: float
        The neutral mass of the ion
    intensity: float
        The intensity of the ion
    charge: int
        The charge of the ion
    mz: float
        The calculated m/z of the ion
    '''

    @property
    def mz(self):
        '''The calculated m/z of the ion

        Returns
        -------
        float
        '''
        return mass_charge_ratio(self.neutral_mass, self.charge)


class PrecursorPurityEstimator(object):
    """Estimate the contamination of precursor ions.

    Attributes
    ----------
    lower_extension : float
        An extra interval below the isolation window to search for coisolating ions,
        intended to collect trailing envelope peaks of lower m/z ions.
    default_width : float
        A default isolation window radius to use when an actual isolation window was
        not available.
    """

    def __init__(self, lower_extension=1.5, default_width=1.5):
        self.lower_extension = lower_extension
        self.default_width = default_width

    def _find_default_envelope_for_mz(self, scan, mz, charge):
        tid = peptide_averagine.isotopic_cluster(mz, charge=charge)
        eid = [scan.peak_set.has_peak(p.mz) for p in tid]
        eid = [EnvelopePair(p.mz, p.intensity) if p is not None else EnvelopePair(tid[i].mz, 1)
               for i, p in enumerate(eid)]
        return Envelope(eid)

    def precursor_purity(self, scan, precursor_peak, isolation_window=None, precursor_charge=None):
        """Calculate the purity of an isolation window.

        If the isolation window is not given, :attr:`default_width` will be used.

        Parameters
        ----------
        scan : :class:`Scan`
            The precursor scan
        precursor_peak : :class:`~.DeconvolutedPeak` or :class:`~.FittedPeak`
            The precursor peak that was deconvolved, whose :attr:`~.DeconvolutedPeak.envelope`
            contains the signal assigned to this precursor ion, or a :class:`~.FittedPeak` which
            is believed to be the monoisotopic peak of the precursor ion. If a :class:`~.FittedPeak`
            is used, `precursor_charge` must be provided.
        isolation_window : :class:`~.IsolationWindow`
            The isolation window to search within
        precursor_charge: :class:`int`, optional
            The charge state of the precursor peak, if the charge of the peak is not known

        Returns
        -------
        float:
            The precursor ion purity, a value of 1.0 indicates that all the signal
            within the isolation window belonged to the assigned peak, and as the value
            approaches 0, the less signal in the isolation window belonged to the precursor
            ion.
        """
        peak_set = scan.peak_set
        mz = precursor_peak.mz
        if isolation_window is None:
            lower_bound = mz - self.default_width
            upper_bound = mz + self.default_width
        else:
            lower_bound = isolation_window.lower_bound
            upper_bound = isolation_window.upper_bound
        try:
            envelope = precursor_peak.envelope
        except AttributeError:
            if precursor_charge is None:
                raise ValueError(
                    "If `precursor_peak` has not been deconvoluted, `precursor_charge` cannot be None")
            envelope = self._find_default_envelope_for_mz(
                scan, precursor_peak.mz, precursor_charge)
        assigned = sum([p.intensity for p in envelope])
        total = sum(
            [p.intensity for p in peak_set.between(lower_bound, upper_bound)])
        if total == 0:
            return 0
        purity = max(min((assigned / total), 1.0), 0)
        return purity

    def coisolation(self, scan, precursor_peak, isolation_window=None, relative_intensity_threshold=0.1,
                    ignore_singly_charged=False):
        """Find any deconvolution solutions which may have partially overlapped the isolation window.

        If the isolation window is not given, :attr:`default_width` will be used.

        Parameters
        ----------
        scan : :class:`Scan`
            The precursor scan
        precursor_peak : :class:`~.DeconvolutedPeak`
            The precursor peak that was deconvolved, whose
            :attr:`~.DeconvolutedPeak.envelope` contains the signal assigned
            to this precursor ion.
        isolation_window : :class:`~.IsolationWindow`
            The isolation window to search within
        relative_intensity_threshold: :class:`float`
            The minimum intensity fraction of `precursor_peak` that another peak must exceed
            in order to be considered.
        ignore_singly_charged: :class:`bool`
            Whether or not to omit singly charged peaks around the isolation window

        Returns
        -------
        :class:`list` of :class:`CoIsolation`
            A list of :class:`CoIsolation` instances describing the deconvoluted peaks
            whose envelopes overlap the isolation window and pass the thresholds.
        """
        peak_set = scan.deconvoluted_peak_set
        mz = precursor_peak.mz

        if isolation_window is None:
            lower_bound = mz - self.default_width
            upper_bound = mz + self.default_width
        else:
            lower_bound = isolation_window.lower_bound
            upper_bound = isolation_window.upper_bound
        extended_lower_bound = lower_bound - self.lower_extension

        peaks = peak_set.between(extended_lower_bound,
                                 upper_bound, use_mz=True)
        others = [
            CoIsolation(p.neutral_mass, p.intensity, p.charge)
            for p in peaks if p != precursor_peak and p.intensity > (
                precursor_peak.intensity * relative_intensity_threshold) and
            p.envelope[-1].mz > lower_bound and ((abs(p.charge) != 1 and ignore_singly_charged) or
                                                 not ignore_singly_charged)
        ]
        return others

    def __call__(self, scan, precursor_peak, isolation_window):
        """A convenience wrapper that calls both :meth:`precursor_purity` and
        :meth:`coisolation`.

        Parameters
        ----------
        scan : :class:`Scan`
            The precursor scan
        precursor_peak : :class:`~.DeconvolutedPeak`
            The precursor peak that was deconvolved, whose
            :attr:`~.DeconvolutedPeak.envelope` contains the signal assigned
            to this precursor ion.
        isolation_window : :class:`~.IsolationWindow`
            The isolation window to search within

        Returns
        -------
        purity: :class:`float`
            The return value of :meth:`precursor_purity`
        coisolation: :class:`list` of :class:`CoIsolation`
            The return value of :meth:`coisolation`

        See Also
        --------
        :meth:`precursor_purity`
        :meth:`coisolation`
        """
        purity = self.precursor_purity(scan, precursor_peak, isolation_window)
        coisolation = self.coisolation(scan, precursor_peak, isolation_window)
        return purity, coisolation


def isolation_window_valid(scan):
    """Check if the isolation window and other precursor-related information
    of `scan` are valid.

    Different vendors periodically report precursor peak m/z values that fall
    outside the reported isolation window. This function verifies that the selected
    precursor m/z is within the isolatoin window's bounds.

    This function always returns :const:`True` when the isolation window is missing or
    empty.

    Parameters
    ----------
    scan : :class:`~.ScanBase`
        The scan to test.

    Returns
    -------
    :class:`bool`:
        Whether the isolation window was valid or not
    """
    window = scan.isolation_window
    if window is None:
        if scan.ms_level == 1:
            # An MS1 scan should never have an isolation window
            return True
        else:
            # We have an MSn scan that does not have an isolation
            # window, which may mean the vendor does not support
            # retrieving this information.
            return True
    if window.is_empty():
        # The vendor or the ScanDataSource populated an empty isolation
        # window, so no analysis is possible
        return True
    pinfo = scan.precursor_information
    # Get the precursor peak's mz, preferrably the corrected one if it is
    # available and check if it is contained in the isolation window.
    mz = pinfo.mz
    try:
        if pinfo.extracted_mz:
            mz = pinfo.extracted_mz
    except (ZeroDivisionError, ValueError):
        pass
    if mz in window:
        return True
    return False


def is_isolation_window_empty(scan):
    window = scan.isolation_window
    if window is None:
        if scan.ms_level == 1:
            # An MS1 scan should never have an isolation window
            return False
        else:
            # We have an MSn scan that does not have an isolation
            # window, which may mean the vendor does not support
            # retrieving this information.
            return False
    if window.is_empty():
        # The vendor or the ScanDataSource populated an empty isolation
        # window, so no analysis is possible
        return False
    pinfo = scan.precursor_information
    precursor = pinfo.precursor
    if precursor is None:
        return False
    try:
        arrays = precursor.arrays
        signal = arrays.between_mz(window.lower_bound, window.upper_bound).intensity.sum()
        return signal <= 1
    except AttributeError:
        if precursor.peaks.is_centroid:
            peaks = precursor.peak_set.between(window.lower_bound, window.upper_bound)
            return sum([p.intensity for p in peaks]) <= 1
        elif precursor.peaks.is_deconvoluted:
            peaks = precursor.deconvoluted_peak_set.between(
                window.lower_bound, window.upper_bound, use_mz=True)
            return sum([p.intensity for p in peaks]) <= 1
        else:
            return False



cimport cython

from brainpy._c.isotopic_distribution cimport TheoreticalPeak

cdef double PROTON

cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
cdef void slide(double mz, list peaklist)
cdef dict scale_dict(dict data, double factor)

cdef class Averagine(object):
    cdef:
        public double base_mass
        public dict base_composition

    cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
    cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
    cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)


@cython.final
cdef class TheoreticalIsotopicPattern(object):
    cdef:
        public list peaklist
        public double origin
        public double offset

    @staticmethod
    cdef TheoreticalIsotopicPattern _create(list peaklist, double origin, double offset)

    cpdef TheoreticalIsotopicPattern clone(self)

    cpdef bint _eq(self, object other)

    @cython.final
    cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)

    cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
    cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
    cpdef TheoreticalIsotopicPattern shift(self, double mz)
    cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
    cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
    cpdef double drop_last_peak(self)

    @cython.final
    cpdef double total(self)

    @cython.final
    cpdef TheoreticalIsotopicPattern normalize(self)

    @cython.final
    cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)

    @cython.final
    cdef inline TheoreticalIsotopicPattern clone_shift(self, double mz)

    @cython.final
    cdef inline TheoreticalPeak get(self, ssize_t i)

    @cython.final
    cdef inline size_t get_size(self)

    @cython.final
    cdef inline double get_monoisotopic_mz(self)
    @cython.final
    cdef inline list get_processed_peaks(self)

    @cython.final
    cpdef size_t _basepeak_index(self)

    @cython.final
    cpdef list incremental_truncation(TheoreticalIsotopicPattern self, double threshold)

    cdef TheoreticalIsotopicPattern clone_drop_last(self)


@cython.final
cdef class AveragineCache(object):
    cdef:
        public dict backend
        public Averagine averagine
        public double cache_truncation
        public bint enabled

    cdef double _make_cache_key(self, double mz)

    cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
    cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)


cpdef double isotopic_shift(int charge=*)


'''
ms_deisotope
------------
A deisotoping and charge state deconvolution library for high resolution mass spectra
with integrated support for reading and writing common mass spectrometry data formats.
'''
try:
    import dill
except ImportError:
    pass

from .version import version

from .averagine import (
    Averagine, peptide, glycan, glycopeptide, heparin,
    heparan_sulfate, permethylated_glycan,
    mass_charge_ratio, neutral_mass,
    calculate_mass, isotopic_shift)
from .deconvolution import (
    AveragineDeconvoluter, CompositionListDeconvoluter,
    AveraginePeakDependenceGraphDeconvoluter,
    CompositionListPeakDependenceGraphDeconvoluter,
    deconvolute_peaks)
from .scoring import MSDeconVFitter, PenalizedMSDeconVFitter, DistinctPatternFitter, IsotopicFitRecord
from .peak_set import DeconvolutedPeak, DeconvolutedPeakSet, DeconvolutedPeakSolution
from .processor import ScanProcessor
from .data_source import MzMLLoader, MzXMLLoader, MSFileLoader


def get_include():
    """Get the include path for C extension headers.

    Returns
    -------
    :class:`str`
    """
    import os
    return os.path.join(
        os.path.dirname(__file__),
        "_c")


__all__ = [
    "Averagine", 'peptide', 'glycan', 'glycopeptide', 'heparin', "heparan_sulfate",
    "permethylated_glycan",
    "mass_charge_ratio", "neutral_mass", "isotopic_shift", "calculate_mass",
    "AveragineDeconvoluter", "CompositionListDeconvoluter",
    "AveraginePeakDependenceGraphDeconvoluter", "CompositionListPeakDependenceGraphDeconvoluter",
    "MSDeconVFitter", "PenalizedMSDeconVFitter", "DistinctPatternFitter", "IsotopicFitRecord",
    "DeconvolutedPeak", "DeconvolutedPeakSet", "DeconvolutedPeakSolution",
    "MzMLLoader", "MzXMLLoader", "MSFileLoader", "ScanProcessor",
    "deconvolute_peaks", 'version'
]



import unittest

from ms_deisotope.data_source.metadata import activation


class TestActivationMetadata(unittest.TestCase):
    def test_is_a_relationships(self):
        assert activation.HCD.is_a(activation.CID)
        assert not activation.HCD.is_a(activation.ETD)

    def test_activation_information(self):
        ai = activation.ActivationInformation(activation.HCD, 30)
        assert ai.has_dissociation_type(activation.HCD)
        assert not ai.has_dissociation_type(activation.ETD)

        mai = activation.MultipleActivationInformation([activation.ETD, activation.HCD], [26, 24])
        assert mai.has_dissociation_type(activation.HCD)
        assert mai.has_dissociation_type(activation.ETD)


if __name__ == '__main__':
    unittest.main()



'''A simple data holder describing a sample which was used to generate
a mass spectrometry dataset.
'''
from ms_deisotope.utils import Base, dict_proxy


@dict_proxy("parameters")
class Sample(Base):
    """Describes a sample used to generate a dataset

    Attributes
    ----------
    id : :class:`object`
        A unique identifier across the samples with which to reference this sample
    name : :class:`str`
        A name for the sample
    parameters : :class:`dict`
        A collection of descriptive

    """
    def __init__(self, id, name=None, parameters=None, **kwargs):
        if parameters is None:
            parameters = dict(kwargs)
        else:
            parameters = dict(parameters)
            parameters.update(kwargs)
        self.id = id
        self.name = name
        self.parameters = parameters



import json
from collections import OrderedDict

from ms_deisotope.data_source.common import (
    PrecursorInformation, ChargeNotProvided,
    ActivationInformation, ScanBase)

from .feature_map import NeutralMassIndex

from ms_deisotope.utils import Base
from ms_deisotope.qc.isolation import CoIsolation


class MSRecordBase(Base):
    def __init__(self, scan_time, drift_time=None, **kwargs):
        self.scan_time = scan_time
        self.drift_time = drift_time
        self._extra_keys = list(kwargs.keys())
        self.__dict__.update(kwargs)

    def __getitem__(self, key):
        return getattr(self, key)

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        return not self == other

    def to_dict(self):
        package = {
            "scan_time": self.scan_time,
        }
        if self.drift_time is not None:
            package['drift_time'] = self.drift_time
        for key in self._extra_keys:
            package[key] = self[key]
        return package

    def get(self, key, default=None):
        try:
            return self[key]
        except Exception:
            return default


class MS1Record(MSRecordBase):
    def __init__(self, scan_time=None, product_scan_ids=None, msms_peaks=None, drift_time=None, **kwargs):
        super(MS1Record, self).__init__(scan_time, drift_time, **kwargs)
        self.product_scan_ids = product_scan_ids or []
        self.msms_peaks = msms_peaks or []

    def to_dict(self):
        package = super(MS1Record, self).to_dict()
        package['product_scan_ids'] = self.product_scan_ids
        package['msms_peaks'] = self.msms_peaks
        return package


class MSnRecord(MSRecordBase):
    def __init__(self, scan_time=None, neutral_mass=None, mz=None, intensity=None, charge=None,
                 precursor_scan_id=None, product_scan_id=None, defaulted=None, orphan=None,
                 drift_time=None, coisolation=None, activation=None, **kwargs):
        super(MSnRecord, self).__init__(scan_time, drift_time, **kwargs)
        self.neutral_mass = neutral_mass
        self.mz = mz
        self.intensity = intensity
        self.charge = charge
        self.precursor_scan_id = precursor_scan_id
        self.product_scan_id = product_scan_id
        self.defaulted = defaulted
        self.orphan = orphan
        self.coisolation = [CoIsolation(*c) if c else c for c in (coisolation or [])]
        self.activation = ActivationInformation.from_dict(
            activation) if isinstance(activation, dict) else activation

    def to_dict(self):
        package = super(MSnRecord, self).to_dict()
        package['neutral_mass'] = self.neutral_mass
        package['mz'] = self.mz
        package['intensity'] = self.intensity
        package['charge'] = self.charge
        package['precursor_scan_id'] = self.precursor_scan_id
        package['product_scan_id'] = self.product_scan_id
        package['defaulted'] = self.defaulted
        package['orphan'] = self.orphan
        package['coisolation'] = self.coisolation
        package['activation'] = self.activation.to_dict()
        return package


class ExtendedScanIndex(object):
    SCHEMA_VERSION = "1.1"

    def __init__(self, ms1_ids=None, msn_ids=None, schema_version=None):
        if schema_version is None:
            schema_version = self.SCHEMA_VERSION
        if ms1_ids is None:
            ms1_ids = {}
        if msn_ids is None:
            msn_ids = {}
        self.ms1_ids = OrderedDict(ms1_ids)
        self.msn_ids = OrderedDict(msn_ids)
        self.schema_version = schema_version

        self._index_bind = None
        self._mass_search_index = None

    def get_scan_dict(self, key):
        try:
            return self.ms1_ids[key]
        except KeyError:
            return self.msn_ids[key]

    def __getitem__(self, key):
        return self.get_scan_dict(key)

    def _package_precursor_information(self, product):
        precursor_information = product.precursor_information
        if precursor_information.extracted_neutral_mass != 0:
            charge = precursor_information.extracted_charge
            if charge == ChargeNotProvided:
                charge = 'ChargeNotProvided'
            package = {
                "neutral_mass": precursor_information.extracted_neutral_mass,
                "mz": precursor_information.extracted_mz,
                "intensity": precursor_information.extracted_intensity,
                "charge": charge,
                "precursor_scan_id": precursor_information.precursor_scan_id,
                "product_scan_id": product.id,
                "scan_time": product.scan_time,
                "defaulted": precursor_information.defaulted,
                "orphan": precursor_information.orphan,
                "coisolation": precursor_information.coisolation,
                "activation": product.activation,
            }
        else:
            charge = precursor_information.charge
            if charge == ChargeNotProvided:
                charge = 'ChargeNotProvided'
            package = {
                "neutral_mass": precursor_information.neutral_mass,
                "mz": precursor_information.mz,
                "intensity": precursor_information.intensity,
                "charge": charge,
                "precursor_scan_id": precursor_information.precursor_scan_id,
                "product_scan_id": product.id,
                "scan_time": product.scan_time,
                "defaulted": precursor_information.defaulted,
                "orphan": precursor_information.orphan,
                "coisolation": precursor_information.coisolation,
                "activation": product.activation,
            }
        if product.has_ion_mobility():
            package['drift_time'] = product.drift_time
        return package

    def add_scan(self, scan):
        '''Add ``scan`` to the index.

        Parameters
        ----------
        scan: :class:`~.ScanBase`
        '''
        if scan.ms_level == 1:
            package = {
                "scan_time": scan.scan_time,
                # would be populated if add_scan_bunch were used
                "product_scan_ids": [],
                "msms_peaks": [
                    p.index.neutral_mass for p in scan.deconvoluted_peak_set
                    if p.chosen_for_msms
                ] if scan.deconvoluted_peak_set is not None else [],
            }
            if scan.has_ion_mobility():
                package['drift_time'] = scan.drift_time
            self.ms1_ids[scan.id] = MS1Record(**package)
        else:
            self.msn_ids[scan.id] = MSnRecord(**self._package_precursor_information(scan))

    def add_scan_bunch(self, bunch):
        '''Add each scan object in ``bunch`` to the index.

        Parameters
        ----------
        scan: :class:`~.ScanBunch`
        '''
        if bunch.precursor is not None:
            package = {
                "scan_time": bunch.precursor.scan_time,
                "product_scan_ids": [
                    product.id for product in bunch.products
                ],
                "msms_peaks": [
                    p.index.neutral_mass for p in bunch.precursor.deconvoluted_peak_set
                    if p.chosen_for_msms
                ] if bunch.precursor.deconvoluted_peak_set is not None else [],
            }
            if bunch.precursor.has_ion_mobility():
                package['drift_time'] = bunch.precursor.drift_time
            self.ms1_ids[bunch.precursor.id] = MS1Record(**package)
        for product in bunch.products:
            self.msn_ids[product.id] = MSnRecord(**self._package_precursor_information(product))

    def update_from_reader(self, reader):
        '''Iterate over ``reader``, accumulating scans in the index.

        Parameters
        ----------
        reader: :class:`~.ScanIterator`
        '''
        for bunch in reader:
            if isinstance(bunch, ScanBase):
                self.add_scan(bunch)
            else:
                self.add_scan_bunch(bunch)

    def dump(self, handle):
        '''Serialize the index to JSON.

        Parameters
        ----------
        handle: file-like
            The file-like object to write the index to
        '''
        mapping = {
            "ms1_ids": [(k, v.to_dict()) for k, v in self.ms1_ids.items()],
            "msn_ids": [(k, v.to_dict()) for k, v in self.msn_ids.items()],
            "schema_version": self.schema_version,
        }
        json.dump(mapping, handle)

    serialize = dump

    def merge(self, other):
        '''Combine the indices in ``other`` with those in ``self``,
        return a copy containing both collections' data.

        Parameters
        ----------
        other: :class:`ExtendedScanIndex`

        Returns
        -------
        :class:`ExtendedScanIndex`
        '''
        dup = ExtendedScanIndex(self.ms1_ids, self.msn_ids)
        dup.ms1_ids.update(other.ms1_ids)
        dup.msn_ids.update(other.msn_ids)
        return dup

    @staticmethod
    def index_file_name(name):
        '''Create a standard file name based on source file name ``name``
        for storing the index

        Parameters
        ----------
        name: str
            The path to the source file to create an adjacent index file
            name for.

        Returns
        -------
        str
        '''
        return name + '-idx.json'

    @classmethod
    def load(cls, handle):
        '''Construct a :class:`ExtendedScanIndex` instance from a file object

        Parameters
        ----------
        handle: file-like

        Returns
        -------
        :class:`ExtendedScanIndex`
        '''
        mapping = json.load(handle)
        ms1_ids = mapping.get("ms1_ids", [])
        mapping['ms1_ids'] = [(k, MS1Record(**v)) for k, v in ms1_ids]
        msn_ids = mapping.get("msn_ids", [])
        mapping['msn_ids'] = [(k, MSnRecord(**v)) for k, v in msn_ids]
        return cls(**mapping)

    deserialize = load

    def get_precursor_information(self, bind=None):
        '''Create a list of :class:`~.PrecursorInformation` objects
        from :attr:`msn_ids`'s records.

        Returns
        -------
        list
        '''
        out = []
        for _, info in self.msn_ids.items():
            mz = info['mz']
            neutral_mass = info['neutral_mass']
            charge = info['charge']
            if charge == "ChargeNotProvided":
                charge = ChargeNotProvided
            intensity = info['intensity']
            precursor_scan_id = info['precursor_scan_id']
            product_scan_id = info['product_scan_id']
            pinfo = PrecursorInformation(
                mz, intensity, charge, precursor_scan_id,
                bind, neutral_mass, charge, intensity,
                product_scan_id=product_scan_id,
                coisolation=info.get('coisolation'))
            out.append(pinfo)
        return out

    def _get_mass_search_index(self, bind=None):
        if self._mass_search_index is not None and (bind is self._index_bind or bind is None):
            return self._mass_search_index
        pinfos = self.get_precursor_information(bind)
        index = NeutralMassIndex(pinfos)
        self._index_bind = bind
        self._mass_search_index = index
        return index

    def find_msms_by_precursor_mass(self, neutral_mass, mass_error_tolerance=1e-5, bind=None):
        '''Find all entries in :attr:`msn_ids` which are within ``mass_error_tolerance`` of
        ``neutral_mass``.

        This method is slow because it reconstructs the search vector on each call.

        Returns
        -------
        list
        '''
        index = self._get_mass_search_index(bind)
        out = index.find_all(neutral_mass, mass_error_tolerance)
        return out



# cython: embedsignature =True

cimport cython

from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_GetSlice


cdef class LCMSFeatureMap(object):
    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.mz)

    def __len__(self):
        return len(self.features)

    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.features[i]
        else:
            return [self.features[j] for j in i]

    cpdef LCMSFeature _search(self, double mz, double error_tolerance):
        cdef:
            long i
            LCMSFeature out
        i = binary_search(self.features, mz, error_tolerance)
        if i == -1:
            return None
        else:
            out = <LCMSFeature>PyList_GET_ITEM(self.features, i)
            return out

    def search(self, mz, error_tolerance=2e-5):
        return self._search(mz, error_tolerance)

    cpdef list _find_all(self, double mz, double error_tolerance):
        cdef:
            long loout, hiout
            list result
        search_sweep(self.features, mz, error_tolerance, &loout, &hiout)
        result = <list>PyList_GetSlice(self.features, loout, hiout)
        return result

    def find_all(self, mz, error_tolerance=2e-5):
        return self._find_all(mz, error_tolerance)

    def index_range(self, lo, hi, error_tolerance=2e-5):
        return (
            binary_search_with_flag(
                self.features, lo, error_tolerance)[0][0],
            binary_search_with_flag(
                self.features, hi, error_tolerance)[0][0])

    def between(self, lo, hi, error_tolerance=2e-5):
        lo_ix = binary_search_with_flag(
            self.features, lo, error_tolerance)[0][0]
        hi_ix = binary_search_with_flag(
            self.features, hi, error_tolerance)[0][0]
        if self[lo_ix].mz < lo:
            lo_ix += 1
        if self[hi_ix] > hi:
            hi_ix -= 1
        if hi_ix < 0:
            hi_ix = 0
        if lo_ix > len(self):
            lo_ix = len(self) - 1
        return self[lo_ix:hi_ix]

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))


cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance):
    cdef:
        int n, lo, hi, mid, low_end, high_end
        double err
        LCMSFeature x
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) / 2
        x = <LCMSFeature>PyList_GET_ITEM(array, mid)
        err = (x.get_mz() - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            # Begin Sweep forward
            while i > 0:
                x = array[i]
                err = (x.get_mz() - mz) / mz
                if abs(err) <= error_tolerance:
                    i -= 1
                    continue
                else:
                    break
            low_end = i
            i = mid + 1

            # Begin Sweep backward
            while i < n:
                x = array[i]
                err = (x.get_mz() - mz) / mz
                if abs(err) <= error_tolerance:
                    i += 1
                    continue
                else:
                    break
            high_end = i
            return list(range(low_end, high_end)), True
        elif (hi - lo) == 1:
            return [mid], False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, False


@cython.cdivision(True)
cdef long binary_search(list array, double mz, double error_tolerance):
    cdef:
        int lo, hi, n, mid, i
        double err
        LCMSFeature x
    lo = 0
    n = hi = PyList_GET_SIZE(array)
    while hi != lo:
        mid = (hi + lo) / 2
        x = <LCMSFeature>PyList_GET_ITEM(array, mid)
        err = (x.get_mz() - mz) / mz
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = err
            i = mid - 1
            while i > 0:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i -= 1

            i = mid + 1
            while i < n:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i += 1
            return best_index
        elif (hi - lo) == 1:
            return -1
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


@cython.cdivision(True)
cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout):
    cdef:
        int i, n, lo, hi, mid, start, end
        double err
        LCMSFeature x
    lo = 0
    n = hi = PyList_GET_SIZE(array)
    loout[0] = 0
    hiout[0] = 0
    while hi != lo:
        mid = (hi + lo) / 2
        x = <LCMSFeature>PyList_GET_ITEM(array, mid)
        err = (x.get_mz() - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            while i > 0:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err > error_tolerance:
                    break
                i -= 1
            start = i + 1
            i = mid + 1
            while i < n:
                x = <LCMSFeature>PyList_GET_ITEM(array, i)
                err = abs((x.get_mz() - mz) / mz)
                if err > error_tolerance:
                    break
                i += 1
            end = i
            loout[0] = start
            hiout[0] = end
            return
        elif (hi - lo) == 1:
            return
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return





cimport cython

import numpy as np
cimport numpy as np

from ms_peak_picker._c.peak_index cimport PeakIndex
from ms_peak_picker._c.peak_set cimport PeakSet
from ms_deisotope._c.peak_set cimport DeconvolutedPeakSet, DeconvolutedPeakSetIndexed, PeakBase


cdef double ppm_error(double x, double y)
cdef double ppm2da(double mass, double error_tolerance)
cdef double da2ppm(double mass, double error_tolerance)


cpdef intensity_getter(peak)
cpdef mz_getter(peak)
cpdef mass_getter(peak)


cpdef list decode_envelopes(np.ndarray[np.float32_t, ndim=1] array)
cpdef DeconvolutedPeakSetIndexed deserialize_deconvoluted_peak_set(dict scan_dict)
cpdef DeconvolutedPeakSetIndexed build_deconvoluted_peak_set_from_arrays(np.ndarray[double, ndim=1] mz_array,
                                                                         np.ndarray[double, ndim=1] intensity_array,
                                                                         np.ndarray[long, ndim=1] charge_array)


ctypedef fused peak_collection:
    PeakSet
    PeakIndex
    DeconvolutedPeakSet
    object


cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1
cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks)


from ms_deisotope._c.feature_map.lcms_feature cimport FeatureBase, LCMSFeatureTreeNode
from ms_deisotope._c.peak_set cimport DeconvolutedPeak
from ms_deisotope._c.averagine cimport TheoreticalIsotopicPattern

cimport numpy as np
import numpy as np

cdef class map_coord:
    cdef:
        public double mz
        public double time


cdef class LCMSFeatureSetFit(object):
    cdef:
        public list features
        public TheoreticalIsotopicPattern theoretical
        public list supporters
        public double score
        public double mz
        public double neutral_mass
        public int charge
        public object data
        public size_t missing_features
        public size_t n_points
        public FeatureBase monoisotopic_feature
        public np.ndarray scores
        public np.ndarray times

    cpdef bint _eq(self, LCMSFeatureSetFit other)
    cpdef bint _ne(self, LCMSFeatureSetFit other)
    cpdef bint _lt(self, LCMSFeatureSetFit other)
    cpdef bint _gt(self, LCMSFeatureSetFit other)


    @staticmethod
    cdef LCMSFeatureSetFit _create(list features, TheoreticalIsotopicPattern theoretical,
                                   double score, int charge, size_t missing_features,
                                   list supporters, object data, double neutral_mass,
                                   size_t n_points, np.ndarray scores, np.ndarray times)



import ms_deisotope
from ms_deisotope.task import TaskBase

from .output import ThreadedMzMLScanStorageHandler, NullScanStorageHandler
from .scan_generator import ScanGenerator


class ScanSink(object):
    def __init__(self, scan_generator, storage_type=NullScanStorageHandler):
        self.scan_generator = scan_generator
        self.scan_store = None
        self._scan_store_type = storage_type

    @property
    def scan_source(self):
        try:
            return self.scan_generator.scan_source
        except AttributeError:
            return None

    @property
    def sample_run(self):
        try:
            return self.scan_store.sample_run
        except AttributeError:
            return None

    def configure_storage(self, storage_path=None, name=None, source=None):
        self.scan_store = self._scan_store_type.configure_storage(
            storage_path, name, source)

    def configure_iteration(self, *args, **kwargs):
        self.scan_generator.configure_iteration(*args, **kwargs)

    def store_scan(self, scan):
        if self.scan_store is not None:
            self.scan_store.accumulate(scan)

    def commit(self):
        if self.scan_store is not None:
            self.scan_store.commit()

    def complete(self):
        if self.scan_store is not None:
            self.scan_store.complete()
        self.scan_generator.close()

    def next_scan(self):
        scan = next(self.scan_generator)
        self.store_scan(scan)
        while scan.ms_level != 1:
            scan = next(self.scan_generator)
            self.store_scan(scan)
        return scan

    def __iter__(self):
        return self

    def __next__(self):
        return self.next_scan()

    def next(self):
        return self.next_scan()


class SampleConsumer(TaskBase):
    MS1_ISOTOPIC_PATTERN_WIDTH = 0.95
    MS1_IGNORE_BELOW = 0.05
    MSN_ISOTOPIC_PATTERN_WIDTH = 0.80
    MSN_IGNORE_BELOW = 0.05

    MS1_SCORE_THRESHOLD = 20.0
    MSN_SCORE_THRESHOLD = 10.0

    def __init__(self, ms_file,
                 ms1_peak_picking_args=None, msn_peak_picking_args=None, ms1_deconvolution_args=None,
                 msn_deconvolution_args=None, start_scan_id=None, end_scan_id=None, storage_path=None,
                 sample_name=None, storage_type=None, n_processes=5,
                 extract_only_tandem_envelopes=False, ignore_tandem_scans=False,
                 ms1_averaging=0, deconvolute=True, verbose=False):

        if storage_type is None:
            storage_type = ThreadedMzMLScanStorageHandler

        self.ms_file = ms_file
        self.storage_path = storage_path
        self.sample_name = sample_name

        self.n_processes = n_processes
        self.storage_type = storage_type
        self.extract_only_tandem_envelopes = extract_only_tandem_envelopes
        self.ignore_tandem_scans = ignore_tandem_scans
        self.ms1_averaging = ms1_averaging

        # for display purposes only
        self.ms1_processing_args = {
            "peak_picking": ms1_peak_picking_args,
        }
        self.msn_processing_args = {
            "peak_picking": msn_peak_picking_args,
        }

        self.deconvolute = deconvolute

        if deconvolute:
            self.ms1_processing_args["deconvolution"] = ms1_deconvolution_args
            self.msn_processing_args["deconvolution"] = msn_deconvolution_args

        n_helpers = max(self.n_processes - 1, 0)
        self.scan_generator = ScanGenerator(
            ms_file,
            number_of_helpers=n_helpers,
            ms1_peak_picking_args=ms1_peak_picking_args,
            msn_peak_picking_args=msn_peak_picking_args,
            ms1_deconvolution_args=ms1_deconvolution_args,
            msn_deconvolution_args=msn_deconvolution_args,
            extract_only_tandem_envelopes=extract_only_tandem_envelopes,
            ignore_tandem_scans=ignore_tandem_scans,
            ms1_averaging=ms1_averaging, deconvolute=deconvolute,
            verbose=verbose)

        self.start_scan_id = start_scan_id
        self.end_scan_id = end_scan_id

        self.sample_run = None

    @classmethod
    def default_processing_configuration(cls, averagine=ms_deisotope.peptide, msn_averagine=None):
        if msn_averagine is None:
            msn_averagine = averagine

        ms1_peak_picking_args = {
            "transforms": [
            ]
        }

        ms1_deconvolution_args = {
            "scorer": ms_deisotope.scoring.PenalizedMSDeconVFitter(20, 2.),
            "max_missed_peaks": 3,
            "averagine": averagine,
            "truncate_after": cls.MS1_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": cls.MS1_IGNORE_BELOW,
            "deconvoluter_type": ms_deisotope.AveraginePeakDependenceGraphDeconvoluter,
            "use_quick_charge": True,
        }

        msn_peak_picking_args = {}

        msn_deconvolution_args = {
            "scorer": ms_deisotope.scoring.MSDeconVFitter(10),
            "averagine": msn_averagine,
            "max_missed_peaks": 1,
            "truncate_after": cls.MSN_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": cls.MSN_IGNORE_BELOW,
            "deconvoluter_type": ms_deisotope.AveraginePeakDependenceGraphDeconvoluter,
            "use_quick_charge": True,
        }

        return (ms1_peak_picking_args, msn_peak_picking_args,
                ms1_deconvolution_args, msn_deconvolution_args)

    def run(self):
        self.log("Initializing Generator")
        self.scan_generator.configure_iteration(self.start_scan_id, self.end_scan_id)
        self.log("Setting Sink")
        sink = ScanSink(self.scan_generator, self.storage_type)
        self.log("Initializing Storage")
        sink.configure_storage(self.storage_path, self.sample_name, self.scan_generator)

        self.log("Begin Processing")
        last_scan_time = 0
        last_scan_index = 0
        i = 0
        for scan in sink:
            i += 1
            if (scan.scan_time - last_scan_time > 1.0) or (i % 1000 == 0):
                self.log("Processed %s (time: %f)" % (
                    scan.id, scan.scan_time,))
                if last_scan_index != 0:
                    self.log("Count Since Last Log: %d" % (scan.index - last_scan_index,))
                last_scan_time = scan.scan_time
                last_scan_index = scan.index
        self.log("Finished Recieving Scans")
        sink.complete()
        self.log("Completed Sample %s" % (self.sample_name,))
        sink.commit()



import os

try:
    from collections import Sequence
except ImportError:
    from collections.abc import Sequence


from .infer_type import MSFileLoader
from .scan import ScanProxyContext


class SubsequenceMappingProxy(object):
    '''Poor man's radix tree.
    '''

    def __init__(self, mapping):
        self.mapping = mapping

    def _resolve(self, key):
        if key in self.mapping:
            return self.mapping[key]
        for k, v in self.mapping.items():
            if not isinstance(k, Sequence):
                continue
            if key in k:
                return v
        best_substr = None
        best_substr_len = -1
        if not isinstance(key, Sequence):
            raise KeyError(key)
        for k, v in self.mapping.items():
            if k in key:
                substr_len = len(k)
                if substr_len > best_substr_len:
                    best_substr = k
                    best_substr_len = substr_len
        if best_substr_len > 0:
            return self.mapping[best_substr]
        raise KeyError(key)

    def __getitem__(self, key):
        return self._resolve(key)

    def __contains__(self, key):
        try:
            found = self[key]
            return found is not None
        except KeyError:
            return False

    def __iter__(self):
        return iter(self.mapping)


class DynamicallyLoadingResolver(object):
    '''Dynamically open data files as they are encountered
    '''

    def __init__(self, root_path=".", mapping=None, opener=None):
        if mapping is None:
            mapping = dict()
        if opener is None:
            opener = self._default_opener
        self.root_path = root_path
        self.mapping = mapping
        self.opener = opener

    def _default_opener(self, path):
        return MSFileLoader(path)

    def _resolve(self, path):
        if path in self.mapping:
            return self.mapping[path]
        _path = os.path.join(self.root_path, path)
        if not os.path.exists(_path):
            raise KeyError(path)
        opened = self.opener(_path)
        self.mapping[path] = opened
        return opened

    def __getitem__(self, key):
        return self._resolve(key)

    def __contains__(self, key):
        try:
            found = self[key]
            return found is not None
        except KeyError:
            return False


class DynamicallyLoadingProxyResolver(DynamicallyLoadingResolver):

    def _default_opener(self, path):
        reader = super(DynamicallyLoadingProxyResolver, self)._default_opener(path)
        return ScanProxyContext(reader)



import os
import gzip
import pickle
import sys

try:
    import faulthandler
    faulthandler.enable()
except ImportError:
    pass

data_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "test_data"))


def datafile(name):
    return os.path.join(data_path, name)


def gzload(path):
    with gzip.open(path, 'rb') as fh:
        if sys.version_info.major > 2:
            return pickle.load(fh, encoding='latin1')
        else:
            return pickle.load(fh)


def example_scan_bunch():
    import ms_deisotope
    reader = ms_deisotope.MSFileLoader(
        datafile("20150710_3um_AGP_001_29_30.mzML.gz"))
    return reader.next()



# -*- coding: utf-8 -*-
'''A base class for deconvolution strategies.
'''
import operator
import logging

from ms_peak_picker import FittedPeak


from ms_deisotope.scoring import IsotopicFitRecord
from ms_deisotope.constants import (
    ERROR_TOLERANCE)

from ms_deisotope.utils import (
    Base)

from .utils import (
    isotopic_shift, drop_placeholders)


logger = logging.getLogger("deconvolution")
info = logger.info
debug = logger.debug
error = logger.error


class DeconvoluterBase(Base):
    """Base class for all Deconvoluter types. Provides basic configuration for common operations,
    regardless of implementation. Because these methods form the backbone of all deconvolution algorithms,
    this class has a C-extension implementation as well.

    Attributes
    ----------
    peaklist : ms_peak_picker.PeakSet
        The centroided mass spectrum to deconvolute
    scorer : IsotopicFitterBase
        The criterion for evaluating individual isotopic pattern fits
    merge_isobaric_peaks : bool
        If multiple passes produce peaks with identical mass values,
        should those peaks be summed
    minimum_intensity : float
        Experimental peaks whose intensity is below this level will be ignored
        by peak querying methods
    scale_method : str
        The name of the method to use to scale theoretical isotopic pattern intensities
        to match the experimental isotopic pattern. For a description of options, see
        :meth:`~.TheoreticalIsotopicPattern.scale`. The default method is `"sum"`.
    use_subtraction : bool
        Whether or not to apply a subtraction procedure to experimental peaks after they
        have been fitted. This is only necessary if the same signal may be examined multiple
        times as in a multi-pass method or when peak dependence is not considered
    verbose : bool
        Produce extra logging information
    """
    use_subtraction = False
    scale_method = 'sum'
    merge_isobaric_peaks = True
    minimum_intensity = 5.
    verbose = False
    peaklist = None
    scorer = None

    def __init__(self, use_subtraction=False, scale_method="sum", merge_isobaric_peaks=True,
                 minimum_intensity=5., *args, **kwargs):
        self.use_subtraction = use_subtraction
        self.scale_method = scale_method
        self.merge_isobaric_peaks = merge_isobaric_peaks
        self.minimum_intensity = minimum_intensity
        self._slice_cache = {}

    def has_peak(self, mz, error_tolerance):
        """Query :attr:`peaklist` for a peak at `mz` within `error_tolerance` ppm. If a peak
        is not found, this method returns a placeholder peak.

        Parameters
        ----------
        mz : float
            The m/z to search for a peak at
        error_tolerance : float
            The parts-per-million error tolerance to search with

        Returns
        -------
        FittedPeak
            A peak from :attr:`peaklist` if present, else a placeholder peak.
        """
        peak = self.peaklist.has_peak(mz, error_tolerance)
        if peak is None or peak.intensity < self.minimum_intensity:
            return FittedPeak(mz, 1.0, 1.0, -1, 0, 0, 0)
        return peak

    def between(self, m1, m2):
        """Take a slice from :attr:`peaklist` using it's :meth:`between` method. Caches
        repeated queries in :attr:`_scan_cache`.

        Parameters
        ----------
        m1 : float
            The low m/z to slice from
        m2 : float
            The high m/z to slice from

        Returns
        -------
        PeakSet
            The subset of peaks from :attr:`peaklist` between `m1` and `m2`
        """
        key = (m1, m2)
        if key in self._slice_cache:
            return self._slice_cache[key]
        else:
            region = self.peaklist.between(m1, m2)
            self._slice_cache[key] = region
            return region

    def match_theoretical_isotopic_distribution(self, theoretical_distribution, error_tolerance=ERROR_TOLERANCE):
        """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
        ppm error. If no experimental peak is found, a placeholder will be used in its stead.

        Parameters
        ----------
        theoretical_distribution : list of :class:`~.TheoreticalPeak`
            The theoretical isotopic pattern to match
        error_tolerance : float, optional
            Parts-per-million error tolerance to permit in searching for matches

        Returns
        -------
        list of :class:`~.FittedPeak`
            The list of matched peaks
        """
        experimental_distribution = [self.has_peak(
            p.mz, error_tolerance) for p in theoretical_distribution]
        return experimental_distribution

    def scale_theoretical_distribution(self, theoretical_distribution, experimental_distribution):
        """Scale up a theoretical isotopic pattern such that its total intensity matches the experimental
        isotopic pattern. This mutates `theoretical_distribution`.

        This assumes that the sum of intensities in `theoretical_distribution` is 1.0. This method is also
        not particularly kind to isotopic patterns where there are missing peaks.

        The scaling algorithm used is controlled by :attr:`scale_method`. For a description of options, see
        :meth:`~.TheoreticalIsotopicPattern.scale`.

        Parameters
        ----------
        theoretical_distribution : :class:`~.TheoreticalIsotopicPattern`
            The theoretical isotopic pattern to scale up
        experimental_distribution : list of :class:`~.FittedPeak`
            The experimental isotopic pattern to use as a reference

        Returns
        -------
        list of :class:`~.TheoreticalPeak`
        """
        theoretical_distribution.scale(
            experimental_distribution, self.scale_method)

    def _evaluate_theoretical_distribution(self, experimental, theoretical, peak, charge):
        """Evaluate a provided theoretical isotopic pattern fit against a
        set of matched experimental peaks.

        Parameters
        ----------
        experimental : list
            A list of experimental fitted peaks to test
        theoretical : :class:`~.TheoreticalIsotopicPattern`
            A list of theoretical isotopic peaks to test
        peak : :class:`~.FittedPeak`
            The seed peak for the isotopic pattern fit
        charge : int
            The target charge state

        Returns
        -------
        :class:`~.IsotopicFitRecord`
        """
        self.scale_theoretical_distribution(theoretical, experimental)
        score = self.scorer(self.peaklist, experimental, theoretical)  # pylint: disable=not-callable
        return IsotopicFitRecord(peak, score, charge, theoretical, experimental)

    def fit_incremental_truncation(self, seed_fit, lower_bound):
        """Fit incrementally truncated versions of the seed fit to check to see if a narrower
        theoretical fit matches the data better.

        Parameters
        ----------
        seed_fit : :class:`~.IsotopicFitRecord`
            The original fit to explore truncations of
        lower_bound : float
            The percentage of total signal remaining to stop truncating at.

        Returns
        -------
        :class:`list` of :class:`~.IsotopicFitRecord`
        """
        patterns = seed_fit.theoretical.incremental_truncation(lower_bound)
        fits = [seed_fit]
        for pattern in patterns[1:]:
            k = len(pattern)
            eid = seed_fit.experimental[:k]
            fit = self._evaluate_theoretical_distribution(
                eid, pattern, seed_fit.seed_peak, seed_fit.charge)
            fits.append(fit)
        return fits

    def subtraction(self, isotopic_cluster, error_tolerance=ERROR_TOLERANCE):
        """Subtract signal attributed to `isotopic_cluster` from the equivalent
        peaks in :attr:`peaklist`, mutating the peaks within.

        This will change the intensity value of the matched `FittedPeak` instances,
        and this is reflected in all of there references.

        Parameters
        ----------
        isotopic_cluster : list of :class:`~.TheoreticalPeak`
            The isotopic cluster to subtract
        error_tolerance : float, optional
            Parts-per-million mass accuracy error tolerance to permit when
            finding matches for `isotopic_cluster`
        """
        for peak in isotopic_cluster:
            match = self.peaklist.has_peak(peak.mz, error_tolerance)
            if match is not None:
                existing = match.intensity
                match.intensity -= peak.intensity
                if (match.intensity < 0) or (peak.intensity > (existing * 0.7)):
                    match.intensity = 1.

    def _merge_peaks(self, peak_list):
        peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))
        if not peak_list:
            return []
        current_peak = peak_list[0]
        merged_peaks = []
        for peak in peak_list[1:]:
            if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:
                current_peak.intensity += peak.intensity
            else:
                merged_peaks.append(current_peak)
                current_peak = peak
        merged_peaks.append(current_peak)
        return merged_peaks

    def _find_next_putative_peak(self, mz, charge, step=1, tolerance=ERROR_TOLERANCE):
        """
        Recalibrates the current peak location given the position of the **next** putative peak
        in a theoretical isotopic cluster.

        Suppose that the peak at `mz` is roughly in the neighborhood of a real isotopic peak,
        but the alignment is bad, so it won't make a good starting point for the search for the
        rest of the peaks in its cluster under a stringent error tolerance.

        However, if we're willing to search for the **next** putative peak with a more permissive error
        tolerance, which we expect will be properly aligned with the rest of its isotopic cluster,
        we can recalibrate the proper starting peak's mz and use that for isotopic cluster fitting.

        Parameters
        ----------
        mz : float
            Starting m/z value to search from
        charge : int
            Charge state to use when calculating the step size in m/z
        step : int, optional
            The number of steps into the putative isotopic cluster to take. Defaults to 1
        tolerance : float, optional
            The error tolerance to accept for finding supporting peaks.

        Returns
        -------
        list
        """
        shift = isotopic_shift(charge)
        next_peak = mz + (shift * step)
        peaklist_slice = self.between(
            next_peak - (next_peak * tolerance),
            next_peak + (next_peak * tolerance))
        candidates = []
        for forward in peaklist_slice:
            prev_peak_mz = forward.mz - (shift * step)
            dummy_peak = FittedPeak(prev_peak_mz, 1.0, 1.0, -1, 0, 0, 0)
            candidates.append((dummy_peak, charge))
        return candidates

    def _find_previous_putative_peak(self, mz, charge, step=1, tolerance=ERROR_TOLERANCE):
        """
        Recalibrates the current peak location given the position of the **previous** putative peak
        in a theoretical isotopic cluster.

        Suppose that the peak at `mz` is roughly in the neighborhood of a real isotopic peak,
        but the alignment is bad, so it won't make a good starting point for the search for the
        rest of the peaks in its cluster under a stringent error tolerance.

        However, if we're willing to search for the **previous** putative peak with a more permissive error
        tolerance, which we expect will be properly aligned with the rest of its isotopic cluster,
        we can recalibrate the proper starting peak's mz and use that for isotopic cluster fitting.

        Parameters
        ----------
        mz : float
            Starting m/z value to search from
        charge : int
            Charge state to use when calculating the step size in m/z
        step : int, optional
            The number of steps into the putative isotopic cluster to take. Defaults to 1
        tolerance : float, optional
            The error tolerance to accept for finding supporting peaks.

        Returns
        -------
        list
        """
        shift = isotopic_shift(charge)
        prev_peak = mz - (shift)
        peaklist_slice = self.between(
            prev_peak - (prev_peak * tolerance),
            prev_peak + (prev_peak * tolerance))
        candidates = []
        for backward in peaklist_slice:
            prev_peak_mz = backward.mz
            if step == 1:
                candidates.extend(self._find_next_putative_peak(
                    prev_peak_mz, charge, 1, tolerance))
            else:
                candidates.extend(
                    self._find_previous_putative_peak(prev_peak_mz, charge, step - 1, tolerance))
        return candidates

    def _check_fit(self, fit):
        if len(drop_placeholders(fit.experimental)) == 1 and fit.charge > 1:
            return False
        if self.scorer.reject(fit):
            return False
        return True

    def __repr__(self):
        type_name = self.__class__.__name__
        return "%s(peaklist=%s)" % (type_name, self.peaklist)


try:
    _DeconvoluterBase = DeconvoluterBase
    from ms_deisotope._c.deconvoluter_base import DeconvoluterBase
except ImportError as e:
    print(e)



import os
import glob
import warnings
import logging

from collections import deque

from six import string_types as basestring
from lxml import etree

try:
    log = logging.getLogger(os.path.basename(__file__))
except Exception:
    log = None
from collections import OrderedDict, defaultdict
from weakref import WeakValueDictionary

try:
    WindowsError
except NameError:
    raise ImportError("Platform Not Supported")


try:
    import comtypes
    from comtypes.client import GetModule, CreateObject
except (ImportError, NameError) as e:
    raise ImportError("Could not import comtypes")

import numpy as np

from ms_deisotope.data_source.common import (
    ScanDataSource,
    RandomAccessScanSource,
    Scan,
    ScanBunch,
    PrecursorInformation,
    ActivationInformation,
    IsolationWindow,
    InstrumentInformation,
    ComponentGroup,
    component,
    FileInformation,
    SourceFile,
    ScanAcquisitionInformation,
    ScanEventInformation,
    ScanWindow)

try:
    # Load previously built COM wrapper
    from comtypes.gen import (
        MassSpecDataReader,
        BaseCommon,
        BaseDataAccess)
    DLL_IS_LOADED = True
except (ImportError, TypeError):
    DLL_IS_LOADED = False


_default_paths = []


def _register_dll_dir(search_paths=None):
    if search_paths is None:
        search_paths = []
    search_paths = list(search_paths)
    search_paths.extend(_default_paths)
    for dll_dir in search_paths:
        try:
            GetModule(os.path.join(dll_dir, 'MassSpecDataReader.tlb'))
            GetModule(os.path.join(dll_dir, 'BaseCommon.tlb'))
            GetModule(os.path.join(dll_dir, 'BaseDataAccess.tlb'))
            global DLL_IS_LOADED
            DLL_IS_LOADED = True
            return True
        except Exception:
            continue
    else:
        return False


def register_dll_dir(search_paths=None):
    if search_paths is None:
        search_paths = []
    if isinstance(search_paths, basestring):
        search_paths = [search_paths]
    loaded = _register_dll_dir(search_paths)
    if not loaded:
        log.debug("Could not resolve Agilent-related DLL")
        search_paths.extend(_default_paths)
        msg = '''
        1) The MassSpecDataReader, BaseCommon, BaseDataAccess DLLs/TLBs may not be installed and
           therefore not registered to the COM server.
        2) The MassSpecDataReader, BaseCommon, BaseDataAccess DLLs/TLBs may not be on these paths:
        %s
        ''' % ('\n'.join(search_paths))
        raise ImportError(msg)


device_to_component_group_map = {
    "QTOF": [
        ComponentGroup("analyzer", [component("quadrupole")], 2),
        ComponentGroup("analyzer", [component("quadrupole")], 3),
        ComponentGroup("analyzer", [component("time-of-flight")], 4)
    ],
    "Quadrupole": [
        ComponentGroup("analyzer", [component("quadrupole")], 2),
    ],
    "TandemQuadrupole": [
        ComponentGroup("analyzer", [component("quadrupole")], 2),
        ComponentGroup("analyzer", [component("quadrupole")], 3),
        ComponentGroup("analyzer", [component("quadrupole")], 4)
    ],
    "IonTrap": [
        ComponentGroup("analyzer", [component("iontrap")], 2)
    ],
    "TOF": [
        ComponentGroup("analyzer", [component("time-of-flight")], 2)
    ]

}


polarity_map = {
    1: -1,
    0: 1,
    3: 0,
    2: None
}


ion_mode_map = {
    0: 'Unspecified',
    1: 'Mixed',
    2: 'EI',
    4: 'CI',
    8: 'Maldi',
    16: 'Appi',
    32: 'Apci',
    64: 'ESI',
    128: 'NanoEsi',
    512: 'MsChip',
    1024: 'ICP',
    2048: 'Jetstream'
}

ionization_map = {
    "EI": component("electron ionization"),
    "CI": component("chemical ionization"),
    "ESI": component("electrospray ionization"),
    "NanoEsi": component("nanoelectrospray"),
    "Appi": component('atmospheric pressure photoionization'),
    "Apci": component("atmospheric pressure chemical ionization"),
    "Maldi": component("matrix assisted laser desorption ionization"),
    "MsChip": component("nanoelectrospray"),
    "ICP": component("plasma desorption ionization"),
    "Jetstream": component("nanoelectrospray")
}


inlet_map = {
    "EI": component("direct inlet"),
    "CI": component("direct inlet"),
    "Maldi": component("particle beam"),
    "Appi": component("direct inlet"),
    "Apci": component("direct inlet"),
    "Esi": component("electrospray inlet"),
    "NanoEsi": component("nanospray inlet"),
    "MsChip": component("nanospray inlet"),
    "ICP": component("component(inductively coupled plasma"),
    "JetStream": component("nanospray inlet"),
}


peak_mode_map = {
    'profile': 0,
    'centroid': 1,
    'profilepreferred': 2,
    'centroidpreferred': 3
}

device_type_map = {
    0: 'Unknown',
    1: 'Mixed',
    2: 'Quadrupole',
    3: 'IsocraticPump',
    4: 'TOF',
    5: 'TandemQuadrupole',
    6: 'QTOF',
    10: 'FlourescenceDetector',
    11: 'ThermalConductivityDetector',
    12: 'RefractiveIndexDetector',
    13: 'MultiWavelengthDetector',
    14: 'ElectronCaptureDetector',
    15: 'VariableWavelengthDetector',
    16: 'AnalogDigitalConverter',
    17: 'EvaporativeLightScatteringDetector',
    18: 'GCDetector',
    19: 'FlameIonizationDetector',
    20: 'ALS',
    21: 'WellPlateSampler',
    22: 'MicroWellPlateSampler',
    23: 'DiodeArrayDetector',
    31: 'CANValves',
    32: 'QuaternaryPump',
    33: 'ChipCube',
    34: 'Nanopump',
    40: 'ThermostattedColumnCompartment',
    41: 'CTC',
    42: 'CapillaryPump',
    50: 'IonTrap'
}


scan_type_map = {
    "Unspecified": 0,
    "All": 7951,
    "AllMS": 15,
    "AllMSN": 7936,
    "Scan": 1,
    "SelectedIon": 2,
    "HighResolutionScan": 4,
    "TotalIon": 8,
    "MultipleReaction": 256,
    "ProductIon": 512,
    "PrecursorIon": 1024,
    "NeutralLoss": 2048,
    "NeutralGain": 4096
}


PEAK_MODE = 0


def make_scan_id_string(scan_id):
    return "scanId=%s" % (scan_id,)


class AgilentDScanPtr(object):
    def __init__(self, index):
        self.index = index

    def __repr__(self):
        return "AgilentDScanPtr(%d)" % (self.index,)


class AgilentDDataInterface(ScanDataSource):
    def _get_spectrum_obj(self, scan, peak_mode=PEAK_MODE):
        index = scan.index
        spectrum = self.source.GetSpectrum_8(rowNumber=index, storageType=peak_mode)
        return spectrum

    def _get_scan_record(self, scan):
        index = scan.index
        record = self.source.GetScanRecord(index)
        return record

    def _scan_index(self, scan):
        return scan.index

    def _scan_id(self, scan):
        record = self._get_scan_record(scan)
        return make_scan_id_string(record.ScanId)

    def _scan_title(self, scan):
        return self._scan_id(scan)

    def _scan_arrays(self, scan):
        spectrum = self._get_spectrum_obj(scan)
        return (np.array(spectrum.XArray, dtype=float),
                np.array(spectrum.YArray, dtype=float))

    def _polarity(self, scan):
        record = self._get_scan_record(scan)
        polarity_enum = record.IonPolarity
        polarity = polarity_map.get(polarity_enum)
        if polarity in (0, None):
            warnings.warn("Unknown Scan Polarity: %r" % (polarity,))
        return polarity

    def _scan_time(self, scan):
        record = self._get_scan_record(scan)
        return record.retentionTime

    def _is_profile(self, scan):
        spectrum_obj = self._get_spectrum_obj(scan)
        mode = spectrum_obj.MSStorageMode
        return mode in (0, 2, 3)

    def _ms_level(self, scan):
        record = self._get_scan_record(scan)
        return record.MSLevel

    def _precursor_information(self, scan):
        if self._ms_level(scan) < 2:
            return None
        spectrum_obj = self._get_spectrum_obj(scan)
        precursor_scan_id = make_scan_id_string(spectrum_obj.ParentScanId)
        n, ions = spectrum_obj.GetPrecursorIon()
        if n < 1:
            return None
        mz = ions[0]
        charge, _ = spectrum_obj.GetPrecursorCharge()
        intensity, _ = spectrum_obj.GetPrecursorIntensity()
        return PrecursorInformation(mz, intensity, charge, precursor_scan_id, self)

    def _acquisition_information(self, scan):
        spectrum_obj = self._get_spectrum_obj(scan)
        try:
            low = spectrum_obj.MeasuredMassRange.Start
            high = spectrum_obj.MeasuredMassRange.End
        except Exception:
            arrays = self._scan_arrays(scan)
            mz_array = arrays[0]
            if len(mz_array) != 0:
                low = mz_array.min()
                high = mz_array.max()
            else:
                low = high = 0
        window = ScanWindow(low, high)
        event = ScanEventInformation(
            self._scan_time(scan),
            window_list=[window])
        return ScanAcquisitionInformation("no combination", [event])

    def _activation(self, scan):
        record = self._get_scan_record(scan)
        return ActivationInformation('cid', record.CollisionEnergy)

    def _isolation_window(self, scan):
        if self._ms_level(scan) < 2:
            return None
        spectrum_obj = self._get_spectrum_obj(scan)
        n, ions = spectrum_obj.GetPrecursorIon()
        if n < 1:
            return None
        return IsolationWindow(0, ions[0], 0)

    def _instrument_configuration(self, scan):
        return self._instrument_config[1]


class _AgilentDDirectory(object):
    @staticmethod
    def create_com_object():
        if not DLL_IS_LOADED:
            raise WindowsError("Could not locate Agilent DLLs")
        reader = CreateObject('Agilent.MassSpectrometry.DataAnalysis.MassSpecDataReader')
        return reader

    @staticmethod
    def create_com_object_filter():
        if not DLL_IS_LOADED:
            raise WindowsError("Could not locate Agilent DLLs")
        no_filter = CreateObject('Agilent.MassSpectrometry.DataAnalysis.MsdrPeakFilter')
        return no_filter

    @staticmethod
    def is_valid(path):
        if os.path.exists(path):
            if os.path.isdir(path):
                return os.path.exists(os.path.join(path, "AcqData", "Contents.xml"))
        return False


class _AgilentMethod(object):
    def __init__(self, method_parameters):
        self.parameters = list(method_parameters)

    def __getitem__(self, i):
        return self.parameters[i]

    def __len__(self):
        return len(self.parameters)

    def __iter__(self):
        return iter(self.parameters)

    def __repr__(self):
        return "_AgilentMethod(%d)" % (len(self),)

    def search_by_name(self, name):
        for param in self:
            try:
                if param['Name'].lower() == name.lower():
                    return param
            except (AttributeError, KeyError):
                continue


class _AgilentDMetadataLoader(object):
    def _has_ms1_scans(self):
        return bool(self._scan_types_flags & scan_type_map['Scan'])

    def _has_msn_scans(self):
        return bool(self._scan_types_flags & scan_type_map['ProductIon'])

    def has_msn_scans(self):
        return self._has_msn_scans()

    def has_ms1_scans(self):
        return self._has_ms1_scans()

    def file_description(self):
        fi = FileInformation(contents={}, source_files=[])
        if self._has_ms1_scans():
            fi.add_content("MS1 spectrum")
        if self._has_msn_scans():
            fi.add_content("MSn spectrum")
        basename = os.path.basename
        dirname = os.path.dirname
        file_queue = deque()
        file_queue.extend(glob.glob(os.path.join(self.dirpath, "AcqData", "*")))
        # for source_file in file_queue:
        while file_queue:
            source_file = file_queue.popleft()
            if os.path.isdir(source_file):
                file_queue.extendleft(glob.glob(os.path.join(source_file, "*")))
            else:
                sf = SourceFile(
                    basename(source_file), dirname(source_file),
                    None, *("Agilent MassHunter nativeID format", "Agilent MassHunter format"))
                sf.add_checksum("sha1")
                fi.add_file(sf, check=False)
        return fi

    def _get_instrument_info(self):
        ion_modes_flags = self.source.MSScanFileInformation.IonModes
        ionization = []
        for bit, label in ion_mode_map.items():
            if ion_modes_flags & bit:
                ionization.append(label)
        configs = []
        i = 1
        for ionizer in ionization:
            groups = [ComponentGroup("source", [ionization_map[ionizer], inlet_map[ionizer]], 1)]
            groups.extend(device_to_component_group_map[self.device])
            config = InstrumentInformation(i, groups)
            i += 1
            configs.append(config)
        self._instrument_config = {
            c.id: c for c in configs
        }
        return configs

    def instrument_configuration(self):
        return sorted(self._instrument_config.values(), key=lambda x: x.id)

    def data_processing(self):
        return []

    def _acquisition_method_xml_path(self):
        return os.path.join(self.dirpath, "AcqData", "AcqMethod.xml")

    def _parse_method_xml(self):
        try:
            path = self._acquisition_method_xml_path()
            tree = etree.parse(path)
            nsmap = {"ns": "http://tempuri.org/DataFileReport.xsd"}
            elt = tree.find(".//ns:SCICDevicesXml", namespaces=nsmap)
            method_xml = etree.fromstring(elt.text)
        except (IOError, OSError, ValueError, TypeError) as e:
            print(e)
            self._method = []
            return self._method
        method = list()
        for section in method_xml.iterfind(".//SectionInfo"):
            section_dict = {}
            for child in section:
                name = child.tag
                value = child.text
                section_dict[name] = value
            method.append(section_dict)
        method = _AgilentMethod(method)
        self._method = method
        return method


_ADM = _AgilentDMetadataLoader
_ADD = _AgilentDDirectory


class AgilentDLoader(AgilentDDataInterface, _ADD, RandomAccessScanSource, _ADM):

    def __init__(self, dirpath, **kwargs):
        self.dirpath = dirpath
        self.dirpath = os.path.abspath(self.dirpath)
        self.dirpath = os.path.normpath(self.dirpath)

        self.source = self.create_com_object()
        self.filter = self.create_com_object_filter()

        try:
            self.source.OpenDataFile(self.dirpath)
        except comtypes.COMError as err:
            raise IOError(str(err))
        self._TIC = self.source.GetTIC()
        self.device = self._TIC.DeviceName
        self._n_spectra = self._TIC.TotalDataPoints
        self._scan_types_flags = self.source.MSScanFileInformation.ScanTypes

        self._producer = self._scan_group_iterator()
        self.initialize_scan_cache()
        self._index = self._pack_index()
        self._get_instrument_info()

    def __reduce__(self):
        return self.__class__, (self.dirpath,)

    @property
    def index(self):
        return self._index

    def __len__(self):
        return len(self.index)

    def __repr__(self):
        return "AgilentDLoader(%r)" % (self.dirpath)

    def reset(self):
        self.make_iterator(None)
        self.initialize_scan_cache()

    def close(self):
        # seems to make attempting to re-open the same datafile cause a segfault
        # self.source.CloseDataFile()
        pass

    def _pack_index(self):
        index = OrderedDict()
        for sn in range(self._n_spectra):
            rec = self._get_scan_record(AgilentDScanPtr(sn))
            index[make_scan_id_string(rec.ScanId)] = sn
        return index

    def _make_scan_index_producer(self, start_index=None, start_time=None):
        if start_index is not None:
            return range(start_index, self._n_spectra)
        elif start_time is not None:
            start_index = self._source.ScanNumFromRT(start_time)
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            return range(start_index, self._n_spectra)
        else:
            return range(0, self._n_spectra)

    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id.

        If the scan object is still bound and in memory somewhere,
        a reference to that same object will be returned. Otherwise,
        a new object will be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        index = self._index[scan_id]
        return self.get_scan_by_index(index)

    def get_scan_by_index(self, index):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        scan_number = int(index)
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = AgilentDScanPtr(scan_number)
            scan = Scan(package, self)
            self._cache_scan(scan)
            return scan

    def get_scan_by_time(self, time):
        time_array = self._TIC.XArray
        lo = 0
        hi = self._n_spectra

        if time == float('inf'):
            return self.get_scan_by_index(len(self) - 1)

        best_match = None
        best_error = float('inf')

        while hi != lo:
            mid = (hi + lo) // 2
            scan_time = time_array[mid]
            err = abs(scan_time - time)
            if err < best_error:
                best_error = err
                best_match = mid
            if scan_time == time:
                return self.get_scan_by_index(mid)
            elif (hi - lo) == 1:
                return self.get_scan_by_index(best_match)
            elif scan_time > time:
                hi = mid
            else:
                lo = mid

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        '''Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        '''
        if scan_id is not None:
            scan_number = self.get_scan_by_id(scan_id).index
        elif index is not None:
            scan_number = int(index)
        elif rt is not None:
            scan_number = self.get_scan_by_time(rt).index
        if require_ms1:
            start_index = scan_number
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            scan_number = start_index
        iterator = self._make_scan_index_producer(start_index=scan_number)
        if grouped:
            self._producer = self._scan_group_iterator(iterator)
        else:
            self._producer = self._single_scan_iterator(iterator)
        return self

    def _make_cache_key(self, scan):
        return scan._data.index

    def _single_scan_iterator(self, iterator=None):
        if iterator is None:
            iterator = self._make_scan_index_producer()
        for ix in iterator:
            packed = self.get_scan_by_index(ix)
            self._cache_scan(packed)
            yield packed

    def _scan_group_iterator(self, iterator=None):
        if iterator is None:
            iterator = self._make_scan_index_producer()

        precursor_scan = None
        product_scans = []

        current_level = 1

        for ix in iterator:
            packed = self.get_scan_by_index(ix)
            self._cache_scan(packed)
            if packed.ms_level > 1:
                # inceasing ms level
                if current_level < packed.ms_level:
                    current_level = packed.ms_level
                # decreasing ms level
                elif current_level > packed.ms_level:
                    current_level = packed.ms_level
                product_scans.append(packed)
            elif packed.ms_level == 1:
                if current_level > 1 and precursor_scan is not None:
                    precursor_scan.product_scans = list(product_scans)
                    yield ScanBunch(precursor_scan, product_scans)
                else:
                    if precursor_scan is not None:
                        precursor_scan.product_scans = list(product_scans)
                        yield ScanBunch(precursor_scan, product_scans)
                precursor_scan = packed
                product_scans = []
            else:
                raise Exception("This object is not able to handle MS levels higher than 2")
        if precursor_scan is not None:
            yield ScanBunch(precursor_scan, product_scans)

    def next(self):
        return next(self._producer)
